"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class Seq_Manip_test(unittest.TestCase):
    def test_rc_T(self):
        sequence='AATTGGCCACGACG'
        rc_sequence=egglib.tools.rc(sequence)
        self.assertEqual(rc_sequence, 'CGTCGTGGCCAATT')
        seq = 'AGGCGaCGCGCATATTAGAGACCGCGatccg'
        self.assertEqual(egglib.tools.rc(seq), 'cggatCGCGGTCTCTAATATGCGCGtCGCCT')
        aln = egglib.Align.create([('no name', seq)], egglib.alphabets.DNA)
        self.assertEqual(egglib.tools.rc(aln.get_sequence(0)), 'CGGATCGCGGTCTCTAATATGCGCGTCGCCT')
        aln2 = egglib.Align.create([('no name', seq)], egglib.alphabets.Alphabet('char', 'ACGT', [], case_insensitive=True))
        self.assertEqual(egglib.tools.rc(aln2.get_sequence(0)), 'CGGATCGCGGTCTCTAATATGCGCGTCGCCT')
        aln4 = egglib.Align.create([('no name', ['A', 'CT', 'A', 'A', 'GGG', 'CT', 'A', 'GGG', 'A'])], egglib.alphabets.Alphabet('string', ['A', 'CT', 'GGG'], []))
        self.assertEqual(egglib.tools.rc(aln4.get_sequence(0)), 'TCCCTAGCCCTTAGT')

    def test_rc_E(self):
        sequence='AAROPGGCCACGACG'
        with self.assertRaises(ValueError):
            rc_sequence=egglib.tools.rc(sequence)

        aln3 = egglib.Align.create([('no name', [12, 14, 31, 62, 12, 14])], egglib.alphabets.Alphabet('range', [0, 100], [-1, 0]))
        with self.assertRaises(ValueError):
            egglib.tools.rc(aln3.get_sequence(0))

        aln5 = egglib.Align.create([('no name', ['A', 'CT', 'A', 'A', 'CT', 'A', 'A'])], egglib.alphabets.Alphabet('custom', ['A', 'CT'], []))
        with self.assertRaises(ValueError):
            egglib.tools.rc(aln5.get_sequence(0))

    def test_compare_T(self):
        sequence1='AATTGGCCACGACG'
        sequence2='CGTCGTGGCCAATT'
        sequence3='AATTGGCCACGACG'
        self.assertFalse(egglib.tools.compare(sequence1, sequence2))
        self.assertTrue(egglib.tools.compare(sequence1, sequence3))
        seq1 = egglib.Align.create([('', sequence1)], egglib.alphabets.DNA)[0].sequence
        seq2 = egglib.Align.create([('', sequence2)], egglib.alphabets.DNA)[0].sequence
        seq3 = egglib.Align.create([('', sequence3)], egglib.alphabets.DNA)[0].sequence
        self.assertFalse(egglib.tools.compare(seq1, seq2))
        self.assertTrue(egglib.tools.compare(seq1, seq3))
        seq4 = egglib.Align.create([('', sequence1)], egglib.alphabets.Alphabet('string', ['A', 'C', 'G', 'T'], []))[0].sequence
        seq6 = egglib.Align.create([('', sequence3)], egglib.alphabets.DNA)[0].sequence
        self.assertFalse(egglib.tools.compare(seq4, sequence2))
        self.assertTrue(egglib.tools.compare(seq4, seq6))

    def test_compare_E(self):
        sequence1='AATTGGCCACGACG'
        sequence2='***rz*dzrfer*f'
        with self.assertRaises(ValueError):
            egglib.tools.compare(sequence1, sequence2)
        seq1a = egglib.Align.create([('', sequence1)], egglib.alphabets.DNA)[0].sequence
        seq1b = egglib.Align.create([('', sequence1)], egglib.alphabets.Alphabet('custom', ['A', 'C', 'G', 'T'], ''))[0].sequence
        with self.assertRaises(ValueError):
            egglib.tools.compare(seq1a, seq1b)


    def test_regex_T(self):
        sequence1='AATTGGCCACGACG'
        self.assertEqual(egglib.tools.regex(sequence1),'(AATTGGCCACGACG)')

    def test_regex_E(self):
        sequence1='***rz*dzrfer*f'
        with self.assertRaises(ValueError):
            egglib.tools.regex(sequence1)

    def test_motif_iter_T(self):
        sequence= 'GGTTCA-T-CTAAGAAC-TG-AC--GCATC-CCT---CG-T--C-T-TTCGA-GTA-GCGTCACTTCTG--TGTTAGG-GG-A--CGCAAG-C--\
CGAGTG--CGTGA--GG-GCGGTCGGC-CCGCGAAATTAACTCCA-GG-TGTACG-AT-GGTGCGT--GACCG-TAG--A--AGGAGA-CCCGTC\
GGGGCCTCT-G-TCGCA-CTAG-GCACT-ATCAG--CCCGTGCCGTGGGG-AAAGG-GCAAGACT--ACCTGGAACGACA-ACT-TCGTGGACTGTCT-\
TA--GC-C-TGGT--ATAGGAATG-T-C---CCCAGGCCCGACG-GCAGGTC-GTT-CTAATATGATGTAG'

        motifs=egglib.tools.motif_iter(sequence, 'GTC',mismatches=0, both_strands=False, case_independent=True, only_base=True)
        l_start=[59,116,187,284,338]
        l_stop=[62,119,190,287,341]
        i=0
        motifs = list(motifs)
        self.assertEqual(len(motifs), len(l_start))
        for hit, start, stop in zip(motifs, l_start, l_stop):
            self.assertEqual(hit[0], start)
            self.assertEqual(hit[1], stop)
            self.assertEqual(hit[2], '+')
            self.assertEqual(hit[3], 0)
            i=+1

        aln = egglib.Container.create([('', 'ATGATGATGATG')], alphabet=egglib.alphabets.DNA)
        motifs = egglib.tools.motif_iter(aln[0].sequence, 'CAT', both_strands=True)
        c = 0
        for start, stop, strand, num in motifs:
            self.assertEqual(start, c*3)
            self.assertEqual(stop, c*3+3)
            self.assertEqual(strand, '-')
            self.assertEqual(num, 0)
            c += 1
        self.assertEqual(c, 4)


    def test_motif_iter_E(self):
        sequence= "GGTTCAGTTCTAAGAACGTGAACAAGCATCTCCTAGTCGGTGTCGTGTTCGAT"
        mtf='GGTTCAGTTCTAAGAACGT'
        
        """
        motifs= egglib.tools.motif_iter(sequence, mtf ,mismatches=0, both_strands=False, case_independent=True,only_base=True)
        with self.assertRaises(ValueError):
            for a,b,c,d in motifs :print a,b,c,d
        """
        
        motifs=egglib.tools.motif_iter(sequence,"")
        with self.assertRaises(ValueError):
            for a,b,c,d in motifs: pass

        motifs=egglib.tools.motif_iter(sequence,'fdez5')
        with self.assertRaises(ValueError):
            for a,b,c,d in motifs: pass

    def test_ungap_T(self):
        f_name='sequence.fas'
        aln = egglib.io.from_fasta(str(path / f_name), labels=False, cls=egglib._interface.Align, alphabet=egglib.alphabets.DNA)
        aln_ugp=egglib.tools.ungap(aln, freq=0.2)
        self.assertIsInstance(aln_ugp, egglib.Align)
        self.assertNotEqual(aln.get_sequence(0).string(),aln_ugp.get_sequence(0).string())
        cnt_ugp=egglib.tools.ungap_all(aln, gaps=['-'])
        self.assertIsInstance(cnt_ugp, egglib.Container)
        self.assertNotEqual(aln.get_sequence(0).string(),cnt_ugp.get_sequence(0).string())
        self.assertRegexpMatches(aln.get_sequence(0).string(), '-')
        self.assertNotRegexpMatches(cnt_ugp.get_sequence(0).string(), '-')
        aln_ugp=egglib.tools.ungap(aln, freq=0.0, gaps=['-'])
        self.assertIsInstance(aln_ugp, egglib.Align)
        self.assertNotEqual(aln.get_sequence(0).string(),aln_ugp.get_sequence(0).string())
        self.assertRegexpMatches(aln.get_sequence(0).string(), '-')
        self.assertNotRegexpMatches(aln_ugp.get_sequence(0).string(), '-')
        aln_ugp=egglib.tools.ungap(aln, freq=1.0, gaps=['-'])
        self.assertIsInstance(aln_ugp, egglib.Align)
        self.assertEqual(aln.get_sequence(0).string(),aln_ugp.get_sequence(0).string())

    def test_ungap_E(self):
        f_name='sequence.fas'
        aln = egglib.io.from_fasta(str(path / f_name), labels=False, cls=egglib._interface.Align, alphabet=egglib.alphabets.DNA)
        cnt = egglib.io.from_fasta(str(path / f_name), labels=False, cls=egglib._interface.Container, alphabet=egglib.alphabets.DNA)
        with self.assertRaises(TypeError):
            aln_ugp=egglib.tools.ungap_all(aln, gaps='-')
        with self.assertRaises(ValueError):
            aln_ugp=egglib.tools.ungap_all(aln, gaps=[])
        with self.assertRaises(TypeError):
            aln_ugp=egglib.tools.ungap_all(cnt)
        with self.assertRaises(ValueError):
            aln_ugp=egglib.tools.ungap_all(aln, gaps=['X'])
        with self.assertRaises(ValueError):
            aln_ugp=egglib.tools.ungap(aln, freq=-10, gaps=['-'])
        with self.assertRaises(ValueError):
            aln_ugp=egglib.tools.ungap(aln, freq=10, gaps=['-'])
        aln.reset()
        with self.assertRaises(ValueError):
            aln_ugp=egglib.tools.ungap(aln, freq=0.5)

    def test_ungap(self):
        sequences = [
            ('name1', 'AGTGTAACGCCACCGTG-'),
            ('name2', 'AG-GTAAC-CCAC-GTG-'),
            ('name3', 'AG-GTAA--CCACCGTG-'),
            ('name4', 'AG-GTAA--CCACC-TGC')]

        aln = egglib.Align.create(sequences, alphabet=egglib.alphabets.DNA)
        self.assertEqual(aln.get_sequence(0).string(), 'AGTGTAACGCCACCGTG-')
        self.assertEqual(aln.get_sequence(1).string(), 'AG-GTAAC-CCAC-GTG-')
        self.assertEqual(aln.get_sequence(2).string(), 'AG-GTAA--CCACCGTG-')
        self.assertEqual(aln.get_sequence(3).string(), 'AG-GTAA--CCACC-TGC')

        cds = egglib.tools.to_codons(aln)
        self.assertEqual(cds.get_sequence(0).string(), 'AGTGTAACGCCACCGTG-')
        self.assertEqual(cds.get_sequence(1).string(), 'AG-GTAAC-CCAC-GTG-')
        self.assertEqual(cds.get_sequence(2).string(), 'AG-GTAA--CCACCGTG-')
        self.assertEqual(cds.get_sequence(3).string(), 'AG-GTAA--CCACC-TGC')

        aln2 = egglib.tools.ungap(aln, freq=0.5)
        self.assertEqual(aln2.get_sequence(0).string(), 'AGGTAACCCACCGTG')
        self.assertEqual(aln2.get_sequence(1).string(), 'AGGTAACCCAC-GTG')
        self.assertEqual(aln2.get_sequence(2).string(), 'AGGTAA-CCACCGTG')
        self.assertEqual(aln2.get_sequence(3).string(), 'AGGTAA-CCACC-TG')

        cds2 = egglib.tools.ungap(cds, freq=0.5)
        self.assertEqual(cds2.get_sequence(0).string(), 'GTACCACCG')
        self.assertEqual(cds2.get_sequence(1).string(), 'GTACCAC-G')
        self.assertEqual(cds2.get_sequence(2).string(), 'GTACCACCG')
        self.assertEqual(cds2.get_sequence(3).string(), 'GTACCACC-')

        sequences = [
            ('name1', [ 1, 2, 3, 4, 5, 0, 7, 0, 0]),
            ('name2', [-1, 0, 3, 4,-1, 6, 7, 8, 0]),
            ('name3', [-1, 0, 3, 4, 0, 0, 7, 8, 0]),
            ('name4', [-1, 0, 3, 4, 0, 6, 7, 8, 9])]

        alph = egglib.alphabets.Alphabet('range', [1, 100], [None, 1])
        aln = egglib.Align.create(sequences, alphabet=alph)
        self.assertEqual(list(aln.get_sequence(0)), [ 1, 2, 3, 4, 5, 0, 7, 0, 0])
        self.assertEqual(list(aln.get_sequence(1)), [-1, 0, 3, 4,-1, 6, 7, 8, 0])
        self.assertEqual(list(aln.get_sequence(2)), [-1, 0, 3, 4, 0, 0, 7, 8, 0])
        self.assertEqual(list(aln.get_sequence(3)), [-1, 0, 3, 4, 0, 6, 7, 8, 9])

        with self.assertRaises(ValueError):
            res = egglib.tools.ungap(aln, freq=0.5)

        res = egglib.tools.ungap(aln, freq=0.5, gaps=[0])
        self.assertEqual(list(res.get_sequence(0)), [ 1, 3, 4, 5, 0, 7, 0])
        self.assertEqual(list(res.get_sequence(1)), [-1, 3, 4,-1, 6, 7, 8])
        self.assertEqual(list(res.get_sequence(2)), [-1, 3, 4, 0, 0, 7, 8])
        self.assertEqual(list(res.get_sequence(3)), [-1, 3, 4, 0, 6, 7, 8])

        res = egglib.tools.ungap(aln, freq=0.5, gaps=[0, -1])
        self.assertEqual(list(res.get_sequence(0)), [3, 4, 0, 7, 0])
        self.assertEqual(list(res.get_sequence(1)), [3, 4, 6, 7, 8])
        self.assertEqual(list(res.get_sequence(2)), [3, 4, 0, 7, 8])
        self.assertEqual(list(res.get_sequence(3)), [3, 4, 6, 7, 8])



"""
    Copyright 2023 Stéphane De Mita, Mathieu Siol, Thomas Coudoux

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import click, pkgutil, pathlib, unittest
from .. import test
from ..test.stats import test_cstats

@click.command(no_args_is_help=True)
@click.option('-b', '--base', is_flag=True, help='run top-level tests')
@click.option('-i', '--io', is_flag=True, help='run io module tests')
@click.option('-t', '--tools', is_flag=True, help='run tools module tests')
@click.option('-s', '--stats', is_flag=True, help='run stats module tests')
@click.option('-c', '--coalesce', is_flag=True, help='run coalesce module tests')
@click.option('-w', '--wrappers', is_flag=True, help='run wrappers module tests')
@click.option('-k', '--custom', type=str, multiple=True, default=[], help='add individual test units')
@click.option('-a', '--all', 'all_', is_flag=True, help='equivalent to -bitscw')
@click.option('-m', '--most', is_flag=True, help='equivalent to -bitsc')
@click.option('--muscle3', is_flag=True, help='test muscle3 rather than muscle5 (if both toggled)')
@click.option('--skip-vcf', is_flag=True, help='skip tests of VCF (dependency on HTSlib)')
@click.option('-L', '--list', 'list_', is_flag=True, help='display list of selected test units')
@click.option('-S', '--stop', is_flag=True, help='stop at first failed test')
@click.option('-e', '--error', is_flag=True, help='terminate with an error if tests don\'t pass')
@click.option('-o', '--output', help='file where to save error messages', type=click.File('w'), default=None)
@click.option('-r', '--stats-report', help='validation of statistics against expectations', type=click.File('w'), default=None)
def main(base, io, tools, stats, coalesce, wrappers, custom, all_, most, muscle3, skip_vcf, list_, stop, error, output, stats_report):
    """
    Run test suite

    If the option --list is specified, the list of selected test units
    is diplayed but no tests are actually run. By default (if no
    selection is specified with other options), the full list of test
    units is displayed.
    """

    # launch stats validation
    if stats_report:
        obj = test_cstats.Statistics_test()
        obj.setUp()
        obj._run_test(stats_report)

    # import all test classes from the test subpackage
    unit_mapping = {}
    group_mapping = {}
    for package in pkgutil.walk_packages(test.__path__):
        group_mapping[package.name] = {}
        for module in pkgutil.walk_packages([str(pathlib.Path(test.__path__[0]) / package.name)]):
            if module.ispkg: raise RuntimeError('problem with test package structure: subpackage found')
            m = module.module_finder.find_module(module.name).load_module()
            for n in dir(m):
                if n[-5:] == '_test':
                    cls = getattr(m, n)
                    if not issubclass(cls, unittest.TestCase):
                        raise RuntimeError(f'problem with test package structure: {module.name}.{n} is not a TestCase subclass')
                    name = n[:-5]
                    if name in unit_mapping:
                        raise RuntimeError(f'problem with test package structure: test unit name {name} repeated')
                    unit_mapping[name] = [package.name, cls]
                    group_mapping[package.name][name] = cls
                    
    # activate all module flags is 'all' is switched on
    if most or all_:
        base = True
        io = True
        tools = True
        stats = True
        if all_:
            wrappers = True

    # include whole requested module tests
    tests = {}
    for name, boolean in [('base', base),
                 ('io', io),
                 ('tools', tools),
                 ('stats', stats),
                 ('coalesce', coalesce),
                 ('wrappers', wrappers)]:
        if boolean: tests[name] = group_mapping[name]

    # include individual units
    for unit in custom:
        try:
            package, cls = unit_mapping[unit]
        except KeyError:
            raise RuntimeError(f'invalid test unit: {unit}')
        if package not in tests: tests[package] = {}
        tests[package][unit] = cls

    # exclude VCF if requested (and present)
    if skip_vcf and 'io' in tests and 'VCF' in tests['io']:
        del tests['io']['VCF']

    # choose muscle3 rather than muscle 5 if requested
    if 'wrappers' in tests and 'muscle3' in tests['wrappers'] and 'muscle5' in tests['wrappers']:
        if muscle3: del tests['wrappers']['muscle5']
        else: del tests['wrappers']['muscle3']

    # list
    if list_:
        if len(tests) == 0:
            tests = group_mapping
        for grp, units in tests.items():
            click.echo(click.style(grp, bold=True))
            for name, unit in units.items():
                click.echo(f'    {name}')

    # perform tests
    elif len(tests):
        suite = unittest.TestSuite()
        for package, units in tests.items():
            for name, cls in units.items():
                suite.addTests(
                    unittest.defaultTestLoader.loadTestsFromTestCase(cls))
        runner = unittest.TextTestRunner(stream=None, verbosity=2,
            failfast=stop)
        results = runner.run(suite)

        # report
        click.echo(f'num tests.............{results.testsRun:.>5d}')
        click.echo(f'errors................{len(results.errors):.>5d}')
        click.echo(f'failures..............{len(results.failures):.>5d}')
        click.echo(f'skipped...............{len(results.skipped):.>5d}')
        click.echo(f'expected failures.....{len(results.expectedFailures):.>5d}')
        click.echo(f'unexpected successes..{len(results.unexpectedSuccesses):.>5d}')
        click.echo(f'stop option...........{int(results.failfast):.>5d}')
        click.echo( 'result:................', nl=False)
        if results.wasSuccessful(): msg = 'pass'; c = 'green'
        else: msg = 'fail'; c = 'red'
        click.echo(click.style(msg.rjust(4, '.'), bold=True, fg=c))

        if output:
            for k in ['errors', 'failures', 'skipped', 'expectedFailures', 'unexpectedSuccesses']:
                items = getattr(results, k)
                h = f'{k}: {len(items)}'
                output.write('=' * len(h) + '\n')
                output.write(h + '\n')
                output.write('=' * len(h) + '\n\n')
                for case, message in items:
                    name = str(case)
                    output.write(name + '\n')
                    output.write('-' * len(name) + '\n')
                    output.write(message + '\n')

        if not results.wasSuccessful() and error:
            raise click.ClickException('tests finished with fail status')



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest, egglib, pathlib, tempfile, os, shutil
path = pathlib.Path(__file__).parent / '..' / 'data'

class VCF_test(unittest.TestCase):
    def setUp(self):
        self.d = tempfile.TemporaryDirectory()
        self.path = pathlib.Path(self.d.name)

    def tearDown(self):
        del self.d

    def test_fname(self):
        egglib.io.VCF(str(path / 'b.vcf')) # RuntimeError if htslib is off
        egglib.io.VCF(str(path / 'b.bcf'))
        with self.assertRaisesRegex(OSError, 'cannot open file: .+not\.exist'):
            egglib.io.VCF(str(path / 'not.exist'))
        with self.assertRaisesRegex(ValueError, 'invalid file: .+b\.gff3'):
            egglib.io.VCF(str(path / 'b.gff3'))

    def test_index_default(self): # (only supported on BCF)
        shutil.copyfile(path / 'b.bcf', self.path / 'b.bcf')
        vcf = egglib.io.VCF(str(self.path / 'b.bcf'))
        self.assertFalse(vcf.has_index)
        egglib.io.index_vcf(str(self.path / 'b.bcf'))
        vcf = egglib.io.VCF(str(self.path / 'b.bcf'))
        self.assertTrue(vcf.has_index)

    def test_index_custom(self):
        shutil.copyfile(path / 'b.bcf', self.path / 'b.bcf')
        vcf = egglib.io.VCF(str(self.path / 'b.bcf'))
        self.assertFalse(vcf.has_index)
        egglib.io.index_vcf(str(self.path / 'b.bcf'), str(self.path / 'index'))
        vcf = egglib.io.VCF(str(self.path / 'b.bcf'))
        self.assertFalse(vcf.has_index)
        vcf = egglib.io.VCF(str(self.path / 'b.bcf'), index=str(self.path / 'index'))
        self.assertTrue(vcf.has_index)

    def test_ctor_args(self):
        with self.assertRaisesRegex(ValueError, 'cannot import index from: .+not\.exist'):
            egglib.io.VCF(fname=str(path / 'b.bcf'),
                          index=str(path / 'not.exist'))

        self.assertTrue((path / 'b.gff3').is_file())
        with self.assertRaisesRegex(ValueError, 'cannot import index from: .+b\.gff3'):
            egglib.io.VCF(fname=str(path / 'b.bcf'),
                          index=str(path / 'b.gff3'))

        for ext in 'vcf', 'bcf':
            vcf = egglib.io.VCF(fname=str(path / f'b.{ext}'))
            self.assertEqual(vcf.num_samples, 4)
            self.assertEqual(vcf.get_samples(), ['INDIV1', 'INDIV2', 'INDIV3', 'INDIV4'])

        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), subset=['INDIV1', 'INDIV3', 'INDIV2', 'INDIV4'])
        self.assertEqual(vcf.num_samples, 4)
        self.assertEqual(vcf.get_samples(), ['INDIV1', 'INDIV2', 'INDIV3', 'INDIV4'])

        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), subset=['INDIV2', 'INDIV1', 'INDIV1', 'INDIV4'])
        self.assertEqual(vcf.num_samples, 3)
        self.assertEqual(vcf.get_samples(), ['INDIV1', 'INDIV2', 'INDIV4'])

        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), subset=['INDIV2', 'INDIV4'])
        self.assertEqual(vcf.num_samples, 2)
        self.assertEqual(vcf.get_samples(), ['INDIV2', 'INDIV4'])

        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), subset=[])
        self.assertEqual(vcf.num_samples, 0)
        self.assertEqual(vcf.get_samples(), [])

        with self.assertRaisesRegex(ValueError, 'unknown sample at position 5'):
            vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), subset=['INDIV1', 'INDIV2', 'INDIV3', 'INDIV4', 'INDIV5'])

        with self.assertRaisesRegex(TypeError, 'subset: expect a sequence of strings'):
            vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), subset=1)

        with self.assertRaisesRegex(TypeError, 'subset: expect a sequence of strings'):
            vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), subset=['INDIV1', 1])

    def test_samples(self):
        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'))
        self.assertEqual([vcf.get_sample(i) for i in range(4)], ['INDIV1', 'INDIV2', 'INDIV3', 'INDIV4'])
        with self.assertRaisesRegex(IndexError, 'sample index out of range'):
            vcf.get_sample(4)

    def test_defaults(self):
        for ext in 'vcf', 'bcf':
            vcf = egglib.io.VCF(fname=str(path / f'b.{ext}'))
            self.assertIsNone(vcf.get_id())
            self.assertIsNone(vcf.get_alleles())
            self.assertIsNone(vcf.get_alternate())
            self.assertIsNone(vcf.get_chrom())
            self.assertIsNone(vcf.get_filter())
            self.assertIsNone(vcf.get_formats())
            self.assertIsNone(vcf.get_genotypes())
            self.assertIsNone(vcf.get_infos())
            self.assertIsNone(vcf.get_phased())
            self.assertIsNone(vcf.get_pos())
            self.assertIsNone(vcf.get_quality())
            self.assertIsNone(vcf.get_reference())
            self.assertIsNone(vcf.get_types())
            self.assertFalse(vcf.is_snp())
            self.assertIsNone(vcf.get_errors())
            self.assertIsNone(vcf.get_info('NO.SUCH.TAG'))
            self.assertIsNone(vcf.get_format('NO.SUCH.TAG', 0))

    def compare_values(self, ctrl, v1, v2, idx, k):
        if isinstance(ctrl, list):
            self.assertEqual(len(ctrl), len(v1), msg=f'site index: {idx+1} - {k}')
            self.assertEqual(len(ctrl), len(v2), msg=f'site index: {idx+1} - {k}')
            for i in range(len(ctrl)):
                self.compare_values_i(ctrl[i], v1[i], v2[i], idx, k, tag=f' item #{i+1}')
        else:
            self.compare_values_i(ctrl, v1, v2, idx, k, tag='')

    def compare_values_i(self, ctrl, v1, v2, idx, k, tag):
        if ctrl is None:
            self.assertIsNone(v1, msg=f'site index: {idx+1} - {k}{tag}')
            self.assertIsNone(v2, msg=f'site index: {idx+1} - {k}{tag}')
        elif isinstance(ctrl, float):
            self.assertAlmostEqual(ctrl, v1, msg=f'site index: {idx+1} - {k}', places=6)
            self.assertAlmostEqual(ctrl, v2, msg=f'site index: {idx+1} - {k}', places=6)
        else:
            self.assertEqual(ctrl, v1, msg=f'site index: {idx+1} - {k}')
            self.assertEqual(ctrl, v2, msg=f'site index: {idx+1} - {k}')

    def test_get_info(self):
        ref_infos = [
            {'DP': 100, 'V': [4], 'W': [41, None], 'AA': 'AA', 'BIDON': 'G',
             'ALT':['.'], 'X': 0.2, 'Y': [1.2], 'GOOD': True, 'INT': None},
            {'AA': 'A', 'TRUC': [407, 12]},
            {'AA': 'C', 'P': [4.13]},
            {'AA': 'C', 'ALT': 'G,C'}, # strings can not represented as multiple values
            {'AA': 'G', 'P': [5.3, None, 3.001]},
            {'AA': 'G', 'TRUC': [None, 500400300]},
            {'AA': 'G', 'DP': None},
            {'AA': 'C', 'ALT': '.'}, # string missing values are not recognized
            {'AA': 'CTC', 'TRUC': [20, None]},
            {'AA': 'A', 'TRI': [1, 2], 'ALT': 'C,G,T', 'GOOD': True},
            {'AA': 'A', 'ALT': '.'}
        ]

        for ext in 'vcf', 'bcf':
            vcf = egglib.io.VCF(fname=str(path / f'b.{ext}'))

            for idx, ref in enumerate(ref_infos):
                self.assertTrue(vcf.read(), msg=f'site index: {idx+1}')
                infos = vcf.get_infos()
                self.assertEqual(infos.keys(), ref.keys(), msg=f'site index: {idx+1}')
                with self.assertRaisesRegex(ValueError, 'invalid info key: NOT.EXIST'):
                    vcf.get_info('NOT.EXIST')
                for k, ctrl in ref.items():
                    v1 = infos[k]
                    v2 = vcf.get_info(k)
                    self.compare_values(ctrl, v1, v2, idx, k)
            self.assertFalse(vcf.read())

    def test_get_format(self):
        ref_format = [
            {},
            {'TEST1': [None, 1, None, None]},
            {'TEST2': [[1, 2], [1, None], [None], [1]]},
            {'TEST3': [0.1, 0.2, None, 0.4]},
            {'TEST4': [[None], [0.2], [0.1, 0.2, 0.3, 0.1], [6]]},
            {'TEST5': ['hipidop', 'a string', 'hap', 'hipidop']},
            {},
            {},
            {},
            {'TEST5': ['.', 'nothing', 'not more', 'something!'],
             'TEST1': [702, 703, 704, 705]},
            {}
        ]

        for ext in 'vcf', 'bcf':
            vcf = egglib.io.VCF(fname=str(path / f'b.{ext}'))

            for idx, ref in enumerate(ref_format):
                self.assertTrue(vcf.read(), msg=f'site index: {idx+1}')
                fmts = vcf.get_formats()
                self.assertIsInstance(fmts, list, msg=f'site index: {idx+1}')
                self.assertEqual(len(fmts), 4, msg=f'site index: {idx+1}')
                for idv, fmt in enumerate(fmts):
                    self.assertEqual(fmt.keys(), ref.keys(), msg=f'site index: {idx+1}')
                    for key, ctrl in ref.items():
                        self.compare_values(ctrl[idv], fmt[key], vcf.get_format(key, idv), idx, key)

                with self.assertRaisesRegex(ValueError, 'invalid format key: NOT.EXIST'):
                    vcf.get_format('NOT.EXIST', 0)
                if len(ref) > 0:
                    with self.assertRaisesRegex(IndexError, 'sample index out of range'):
                        vcf.get_format(list(ref)[0], 4)
                    self.assertEqual(vcf.get_format(list(ref)[0], -1),
                                     vcf.get_format(list(ref)[0], 3))
                    self.assertEqual(vcf.get_format(list(ref)[0], -2),
                                     vcf.get_format(list(ref)[0], 2))
                    self.assertEqual(vcf.get_format(list(ref)[0], -3),
                                     vcf.get_format(list(ref)[0], 1))
                    self.assertEqual(vcf.get_format(list(ref)[0], -4),
                                     vcf.get_format(list(ref)[0], 0))
                    with self.assertRaisesRegex(IndexError, 'sample index out of range'):
                        vcf.get_format(list(ref)[0], -5)

            self.assertFalse(vcf.read())

    def test_error(self):
        for ext in 'vcf', 'bcf':
            vcf = egglib.io.VCF(fname=str(path / f'b.{ext}'))
            while vcf.read():
                self.assertEqual(vcf.get_errors(), [])

        s = ('##fileformat=VCFv4.2\n'
             '##contig=<ID=ctg1,len=1000>\n'
             '##INFO=<ID=FLOAT,Number=1,Type=Float,Description="Something">\n'
             '##FORMAT=<ID=INT,Number=1,Type=Integer,Description="Something else">\n'
             '#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	INDIV1	INDIV2	INDIV3	INDIV4\n'
             'ctg1	10	.	A	.	.	PASS	FLOAT=1	INT	1	2	3	4\n'
             'ctg1	20	.	A	.	.	PASS	FLOAT=1.1;DOUBLE=4.5	INT	1	2	3	4\n'
             'ctg1	30	.	A	.	.	PASS	FLOAT=1.1	INT	1	2	XXX	4\n'
             'ctg1	40	.	A	.	.	PASS	FLOAT=1.5	INT	1	2	3	4\n')

        with open(self.path / 'tmp.vcf', 'w') as f:
            f.write(s)

        vcf = egglib.io.VCF(fname=str(self.path / 'tmp.vcf'))
        self.assertTrue(vcf.read())
        self.assertEqual(vcf.get_errors(), [])
        self.assertTrue(vcf.read())
        self.assertGreater(len(vcf.get_errors()), 0)
        with self.assertRaisesRegex(ValueError, 'critical error while reading a variant'):
            vcf.read()
        self.assertTrue(vcf.read())
        self.assertEqual(vcf.get_errors(), [])
        self.assertFalse(vcf.read())

    def helper_test_goto(self, vcf:egglib.io.VCF,
                ctg:str, pos:int = None, limit:str = None,
                expected_ctg_error:bool = False,
                expected_pos_error:bool = False,
                expected_position:int = None):
        if expected_ctg_error:
            with self.assertRaisesRegex(ValueError, f'unknown target name: {ctg}'):
                if pos is None: vcf.goto(ctg)
                elif limit is None: vcf.goto(ctg, pos)
                else: vcf.goto(ctg, pos, limit)
            return
        if expected_pos_error:
            with self.assertRaisesRegex(ValueError, f'position not found: {pos}'):
                if limit is None: vcf.goto(ctg, pos)
                else: vcf.goto(ctg, pos, limit)
            return
        if pos is None: vcf.goto(ctg)
        elif limit is None: vcf.goto(ctg, pos)
        else: vcf.goto(ctg, pos, limit)
        self.assertEqual(vcf.get_chrom(), ctg)
        if expected_position is None and pos is not None:
            expected_position = pos
        if expected_position is not None:
            self.assertEqual(vcf.get_pos(), expected_position)

    def test_goto(self):
        # load without index
        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'))
        with self.assertRaisesRegex(ValueError, 'an index is required'):
            vcf.goto('ctg2')

        # create index and load
        idx = str(path / 'idx')
        egglib.io.index_vcf(str(path / 'b.bcf'), idx)
        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), index=idx)
        self.assertTrue(vcf.has_index)

        # go to a contig
        self.helper_test_goto(vcf, 'ctg2', expected_position=1014)

        # next pos
        self.assertTrue(vcf.read())
        self.assertEqual(vcf.get_chrom(), 'ctg2')
        self.assertEqual(vcf.get_pos(), 1015)

        # go to a position
        self.helper_test_goto(vcf, 'ctg2', 1049)

        # go to an invalid contig
        self.helper_test_goto(vcf, 'ctgN', expected_ctg_error=True)

        # go back near beginning
        self.helper_test_goto(vcf, 'ctg1', 1000)

        # reopen and go to first position of first contig
        vcf = egglib.io.VCF(fname=str(path / 'b.bcf'), index=idx)
        self.helper_test_goto(vcf, 'ctg1', expected_position=999)

        # go to a non-existing position and fix with limit
        self.helper_test_goto(vcf, 'ctg2', 1017, expected_pos_error=True)
        self.helper_test_goto(vcf, 'ctg2', 1017, limit=1019, expected_pos_error=True)
        self.helper_test_goto(vcf, 'ctg2', 1017, limit=1020, expected_position=1019)
        self.helper_test_goto(vcf, 'ctg3', 1067, limit=egglib.io.VCF.END, expected_position=1099)

        with self.assertRaisesRegex(ValueError, '`limit` must be larger than `pos`'):
            self.helper_test_goto(vcf, 'ctg2', 1019, limit=1019)
        with self.assertRaises(TypeError):
            self.helper_test_goto(vcf, 'ctg2', 1019, limit='END')
        with self.assertRaisesRegex(ValueError, '`limit` must be strictly positive'):
            self.helper_test_goto(vcf, 'ctg2', 1019, limit=-10)

        # another test
        self.helper_test_goto(vcf, 'ctg2', 1017, limit=egglib.io.VCF.END, expected_position=1019)

        # go past the end of a contig
        self.helper_test_goto(vcf, 'ctg2', 1200, expected_pos_error=True)
        self.helper_test_goto(vcf, 'ctg2', 1200, limit=1300, expected_pos_error=True)
        self.helper_test_goto(vcf, 'ctg2', 1200, limit=egglib.io.VCF.END, expected_pos_error=True)
        self.assertEqual(vcf.get_info(), None)

        # go back to actual beginning
        self.helper_test_goto(vcf, 'ctg1', expected_position=999)

        # go past the end of the file
        self.helper_test_goto(vcf, 'ctg3', 9999, expected_pos_error=True)
        self.helper_test_goto(vcf, 'ctg3', 9999, limit=10000, expected_pos_error=True)
        self.helper_test_goto(vcf, 'ctg3', 9999, limit=egglib.io.VCF.END, expected_pos_error=True)
        self.assertEqual(vcf.get_info(), None)

##### add methods to test accessors ####################################

accessor_data = {
    'chrom': ['ctg1', 'ctg1', 'ctg1', 'ctg1', 'ctg2', 'ctg2', 'ctg2', 'ctg2', 'ctg2', 'ctg3', 'ctg3'],
    'pos': [999, 1000, 1009, 1010, 1014, 1015, 1019, 1029, 1049, 1059, 1099],
    'id': [['snp1', 'first', 'zero'], ['snp11'], ['snp2'], ['snp21'], [],
           ['snp201', 'snp+'], ['snp3', 'snp3'], ['snp4'], ['snp5'], ['no_snp'], []],
    'reference': ['A', 'A', 'C', 'C', 'G', 'G', 'G', 'C', 'CTC', 'A', 'A'],
    'alternate': [['T'], ['AA'], ['G', 'T'], ['G', 'CTT'], ['TAA'],
                  ['GAA'], ['T'], ['A'], ['ATG'], ['C'], []],
    'alleles': [['A', 'T'], ['A', 'AA'], ['C', 'G', 'T'],
                ['C', 'G', 'CTT'], ['G', 'TAA'], ['G', 'GAA'],
                ['G', 'T'], ['C', 'A'], ['CTC', 'ATG'], ['A', 'C'], ['A']],
    'quality': [4, None, None, None, None, None, None, None, None, None, None],
    'filter': [[], [], ['triple'], ['triple', 'multi'], [], [], [], [], [], [], []],
    'errors': [[], [], [], [], [], [], [], [], [], [], []],
    'phased': [(True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [True], [True]]),
               (True, [[True], [True], [], [True]]),
               (False, [[True, True], [True, False], [False, False], [False, False]]),
               (False, [[True], [True], [False], [True]])],
    'types': [['SNP'], ['INDEL'], ['SNP'], ['SNP', 'INDEL'], ['OTHER'],
              ['INDEL'], ['SNP'], ['SNP'], ['MNP'], ['SNP'], []],
    'is_snp': [True, False, True, False, False, False, True, True, False, True, False],
    'genotypes': [
        [ ['A', 'A'], ['A', 'T'], ['A', 'A'], ['T', 'T'] ],
        [ ['A', 'A'], ['A', 'AA'], ['A', 'A'], ['AA', 'AA'] ],
        [ ['G', 'G'], ['C', 'G'], ['G', 'C'], ['T', 'T'] ],
        [ ['G', 'G'], ['C', 'G'], ['G', 'C'], ['CTT', 'CTT'] ],
        [ ['G', 'TAA'], ['TAA', 'TAA'], ['G', 'G'], ['G', 'G'] ],
        [ ['G', 'GAA'], ['GAA', 'GAA'], ['G', 'G'], ['G', 'G'] ],
        [ ['G', 'G'], ['G', 'T'], ['G', 'G'], ['T', 'T'] ],
        [ ['C', 'C'], ['C', 'A'], ['C', 'C'], ['A', 'A'] ],
        [ ['CTC', 'CTC'], ['CTC', 'ATG'], [None], ['ATG', 'ATG'] ],
        [ ['A', 'A', 'A'], ['A', 'A', 'A'], ['A', 'A', 'C'], ['A', 'C', 'C'] ],
        [ ['A', 'A'], ['A', None], [None, None], [None, None] ]
    ]
}

for what in accessor_data:
    def f(self, what=what):
        attr = what if what == 'is_snp' else f'get_{what}'
        for ext in 'vcf', 'bcf':
            vcf = egglib.io.VCF(fname=str(path / f'b.{ext}'))
            for i, v in enumerate(accessor_data[what]):
                self.assertTrue(vcf.read(), msg=f'read() returned False extension={ext} variant=#{i+1} what={what}')
                val = getattr(vcf, attr)()
                self.assertEqual(val, v,
                    msg=f'extension={ext} variant=#{i+1} what={what} exp={v} received={val}')
            self.assertFalse(vcf.read(), msg=f'extra read() returned True extension={ext} what={what}')
    setattr(VCF_test, f'test_{what}', f)



"""
    Copyright 2018-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from . import eggwrapper as _eggwrapper

class Alphabet(object):
    """
    Define acceptable lists of alleles for a type of data.
    This class is designed to be associated to EggLib objects holding
    diversity data (:class:`~Align`, :class:`~Container`, and :class:`~Site`).
    It defines the lists of exploitable and missing allelic values (both
    being allowed in all objects using this alphabet).
    Several pre-defined instances instances are
    available in :ref:`alphabets`, but the user can easily define his own
    alphabets using this class. It is strongly advised to use the pre-defined
    alphabet designed for DNA sequences as it is optimized for access speed.

    :param cat: Type of the alphabet, as a string taken from the following list:

        +---------------+-------------------------------------+
        |               |                                     |
        +---------------+-------------------------------------+
        | ``int``       | Positive or negative, discrete      |
        |               | integers. ``range`` is more         |
        |               | adapted to cases with many alleles. |
        +---------------+-------------------------------------+
        | ``char``      | Single-character strings            |
        |               | (such as DNA).                      |
        +---------------+-------------------------------------+
        | ``string``    | Strings representing segments of    |
        |               | biological sequences (appropriate   |
        |               | for codons or indel variants).      |
        +---------------+-------------------------------------+
        | ``custom``    | Alleles represented by a code or a  |
        |               | name (e.g. for genomic              |
        |               | rearrangements). In particular,     |
        |               | this type of alphabets does not     |
        |               | allow sequence concatenation,       |
        |               | contrary to ``string``.             |
        +---------------+-------------------------------------+
        | ``range``     | Continuous ranges of integers       |
        |               | (regardless of their sign). This is |
        |               | much more efficient than using an   |
        |               | integer alphabet with many alleles. |
        +---------------+-------------------------------------+

    :param expl: list of exploitable alleles (for ``range``, see :ref:`below <range_spec>`).

    :param miss: list of missing alleles (for ``range``, see :ref:`below <range_spec>`).
        There must not be any overlap between exploitable and missing
        alleles in any case.

    :param case_insensitive: if set to ``True``, case will be ignored
        for allele comparison (alleles differing in case only will be
        considered to be identical). Note that this does not preserve
        case of provided alleles which are all converted to upper case.
        Only accepted for ``char`` and ``string`` alphabets.

    :param name: name of the alphabet. Alphabet name is of relatively
        modest importance and is not required to be unique. The alphabet
        name is used in error messages when an invalid allele is
        processed. By default, use the alphabet type as name.

    .. _range_spec:

    For ``range`` alphabets, both parameters *expl* and *miss*
    must be specified as a ``(start, stop)`` expression. All values from
    ``start`` up to ``stop``-1 will be considered as valid allele values
    (``stop`` is not included).
    To specify a unique value, use ``(value, value+1)``. To specify no
    value at all, use ``None`` (you can also abuse the syntax by using
    ``(x, x)`` with any ``x`` value. You can omit either term, or both,
    to denote a semi-infinite or fully infinite range. Of course, setting
    ``expl=(None, None)`` prevent you from specifying any missing allele
    value.
    """

    @classmethod
    def _make(cls, obj): # make an instance around a provided _eggwrapper.AbstractBaseAlphabet subclass instance
        inst = cls.__new__(cls)
        inst._obj = obj
        return inst

    def __init__(self, cat, expl, miss, case_insensitive=False, name=None):
        # create object of right type
        if cat == 'int':
            if case_insensitive == False: self._obj = _eggwrapper.IntAlphabet()
            else: raise ValueError('case insensitivity not available for int alphabet')
        elif cat == 'char':
            if case_insensitive == False: self._obj = _eggwrapper.CharAlphabet()
            else: self._obj = _eggwrapper.CaseInsensitiveCharAlphabet()
        elif cat == 'string':
            if case_insensitive == False: self._obj = _eggwrapper.StringAlphabet()
            else: self._obj = _eggwrapper.CaseInsensitiveStringAlphabet()
        elif cat == 'custom':
            if case_insensitive == False: self._obj = _eggwrapper.CustomStringAlphabet()
            else: raise ValueError('case insensitivity not available for custom alphabet')
        elif cat == 'range':
            if case_insensitive == False: self._obj = _eggwrapper.RangeAlphabet()
            else: raise ValueError('case insensitivity not available for range alphabet')
        else: raise ValueError('invalid alphabet category')

        # set name
        if name is None:
            name = type(self._obj).__name__
        elif not isinstance(name, str): raise TypeError('name must be a string')
        self._obj.set_name(name)

        # set type
        self._obj.set_type(cat)

        # configure RangeAlphabet
        if cat == 'range':
            if expl is None: expl = [0, 0]
            else:
                try: expl = list(expl)
                except TypeError: raise TypeError('invalid exploitable range definition: {0}'.format(expl))
            if len(expl) == 1: miss = [expl[0], expl[0]+1]
            if len(expl) != 2: raise ValueError('invalid exploitable range specification')
            if expl[0] is None: expl[0] = - _eggwrapper.MAX_ALLELE_RANGE
            elif expl[0] < - _eggwrapper.MAX_ALLELE_RANGE: raise ValueError('lower bound of exploitable range is out of allowed range')
            if expl[1] is None: expl[1] = _eggwrapper.MAX_ALLELE_RANGE + 1
            elif expl[1] > _eggwrapper.MAX_ALLELE_RANGE + 1: raise ValueError('higher bound of exploitable range is out of allowed range')
            if not isinstance(expl[0], int) or not isinstance(expl[1], int): raise ValueError('invalid exploitable range specification')
            if expl[1] < expl[0]: raise ValueError('invalid exploitable range specification')
            self._obj.set_exploitable(expl[0], expl[1])
            if miss is None: miss = [0, 0]
            else:
                try: miss = list(miss)
                except TypeError: raise TypeError('invalid missing range definition: {0}'.format(miss))
            if len(miss) == 1: miss = [miss[0], miss[0]+1]
            if len(miss) != 2: raise ValueError('invalid missing range specification')
            if miss[0] is None: miss[0] = - _eggwrapper.MAX_ALLELE_RANGE
            elif miss[0] < - _eggwrapper.MAX_ALLELE_RANGE: raise ValueError('lower bound of missing range out of allowed range')
            if miss[1] is None: miss[1] = _eggwrapper.MAX_ALLELE_RANGE + 1
            elif miss[1] > _eggwrapper.MAX_ALLELE_RANGE + 1: raise ValueError('higher bound of missing range out of allowed range')
            if not isinstance(miss[0], int) or not isinstance(miss[1], int): raise ValueError('invalid missing range specification')
            if miss[1] < miss[0]: raise ValueError('invalid missing range specification')
            self._obj.set_missing(miss[0], miss[1])
        else:
            if expl is None: expl = []
            if miss is None: miss = []
            for i in expl:
                self._check(i)
                self._obj.add_exploitable(i)
            for i in miss:
                self._check(i)
                self._obj.add_missing(i)

    def _check(self, value):
        # check that the passed allelic value is of the appropriate type for the instance
        if self._obj.get_type() == 'DNA':
            if not (isinstance(value, str) and len(value) == 1): raise TypeError('expect single-character string alleles')
            return
        if self._obj.get_type() == 'codons':
            if not (isinstance(value, str) and len(value) == 3): raise TypeError('expect three-character string alleles')
            return
        if self._obj.get_type() == 'char':
            if not (isinstance(value, str) and len(value) == 1): raise TypeError('expect single-character string alleles')
            return
        if self._obj.get_type() == 'int':
            if not isinstance(value, int): raise TypeError('expect integer alleles')
            return
        if self._obj.get_type() == 'string':
            if not isinstance(value, str): raise TypeError('expect string alleles')
            return
        if self._obj.get_type() == 'custom':
            if not isinstance(value, str): raise TypeError('expect string alleles')
            return
        if self._obj.get_type() == 'range':
            raise ValueError('range alphabets cannot be modified')
            return
        raise TypeError('invalid type for {0} alphabet: {1}'.format(self._obj.get_type(), type(value).__name__))

    @property
    def name(self):
        """
        Name of the alphabet. The value might be changed.
        """
        return self._obj.get_name()

    @name.setter
    def name(self, value):
        if not isinstance(value, str): raise TypeError('name must be a string')
        self._obj.set_name(value)

    def get_alleles(self):
        """
        Generate the list of alleles.
        Return a tuple with two items: the list of exploitable
        and the list of missing alleles, respectively. For a ``range`` alphabet, each
        list is replaced by a tuple with the bounds of the range
        (replaced by ``None`` in case of an empty range).
        """
        if isinstance(self._obj, _eggwrapper.RangeAlphabet):
            a = self._obj.first_exploitable()
            b = self._obj.end_exploitable()
            c = self._obj.first_missing()
            d = self._obj.end_missing()
            if b == a:
                ab = None
            else:
                if a == -_eggwrapper.MAX_ALLELE_RANGE: a = None
                if b ==  _eggwrapper.MAX_ALLELE_RANGE + 1: b = None
                ab = (a, b)
            if d == c:
                cd = None
            else:
                if c == -_eggwrapper.MAX_ALLELE_RANGE: c = None
                if d ==  _eggwrapper.MAX_ALLELE_RANGE + 1: d = None
                cd = (c, d)
            return (ab, cd)
        else:
            return ([self._obj.get_value(i) for i in range(self._obj.num_exploitable())],
                    [self._obj.get_value(-i-1) for i in range(self._obj.num_missing())])

    @property
    def num_alleles(self):
        """
        Total number of alleles.
        """
        return self._obj.num_exploitable() + self._obj.num_missing()

    @property
    def num_exploitable(self):
        """
        Number of exploitable alleles.
        """
        return self._obj.num_exploitable()

    @property
    def num_missing(self):
        """
        Number of missing alleles.
        """
        return self._obj.num_missing()

    def add_exploitable(self, value):
        """
        Add an exploitable allele to the instance. Not allowed for ``range``
        alphabets and for the special alphabet :py:obj:`.alphabets.DNA`.
        The allele value
        must be of the appropriate type, and unique.
        """
        self._check(value)
        self._obj.add_exploitable(value)

    def add_missing(self, value):
        """
        Add a missing allele to the instance. Not allowed for ``range``
        alphabets and for the special alphabet :py:obj:`.alphabets.DNA`.
        The allele value must be of the appropriate type, and unique.
        """
        self._check(value)
        self._obj.add_missing(value)

    @property
    def case_insensitive(self):
        """
        Boolean indicating if the alphabet is case-insensitive.
        """
        return self._obj.case_insensitive()

    @property
    def type(self):
        """ Type of the alphabet. Same as the *cat* argument of the constructor. """
        return self._obj.get_type()

    def get_code(self, value):
        """
        Return the code of a given allele. Missing alleles are indicated
        by a negative code.
        """
        try: return self._obj.get_code(value)
        except TypeError: raise TypeError('invalid allele type for this alphabet')

    def get_value(self, code):
        """
        Return the value of for given code. Missing alleles are indicated
        by a negative code.
        """
        if not isinstance(code, int): raise TypeError('allele codes must be represented by an integer')
        return self._obj.get_value(code)

    def lock(self):
        """
        Lock the alphabet. When the alphabet is locked, a :exc:`ValueError`
        will be raised if an attempt to modify the alphabet is made.
        """
        self._obj.lock()

    @property
    def locked(self):
        """
        ``True`` if the alphabet is locked.
        """
        return self._obj.is_locked()

#
# WARNING: ORDER OF DNA, PROTEIN and CODON ALPHABETS _MUST NOT_ BE MODIFIED
# C++ implementation of GeneticCode depends on it
#

DNA = Alphabet._make(_eggwrapper.get_static_DNAAlphabet())
"""Alphabet optimized for DNA sequences (case-insensitive)."""

codons = Alphabet._make(_eggwrapper.get_static_CodonAlphabet())
"""Alphabet for codon triplets (only upper case)."""

protein = Alphabet('char', 'ACDEFGHIKLMNPQRSTVWY*', '-X?', case_insensitive=False, name='protein')
"""Alphabet for amino acids (only upper case). Stop codons are supported as ``*``."""
protein.lock()

positive_infinite = Alphabet('range', [0, None], [-1], case_insensitive=False, name='positive infinite')
"""Alphabet with all positive integer values."""
positive_infinite.lock()

genepop = Alphabet('range', (1, 1000), (0, 1), name='genepop')
"""Alphabet matching the Genepop format."""
genepop.lock()

binary = Alphabet('int', [0, 1], [-9, -1, 999], name='binary')
"""Alphabet for binary data (0/1 integers)."""
binary.lock()



"""
    Copyright 2016-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import alphabets
from .. import _site
from .. import _freq

class ProbaMisoriented(object):
    """
    Error rate of polymorphism orientation using an outgroup.

    :param align: an :class:`.Align` containing the sites to analyse.
    :param struct: a :class:`.Structure` instance.

    Only sites that are either variable
    within the ingroup or have a fixed difference with respect to the
    outgroup are considered. Sites with more than two different alleles
    in the ingroup, or more than one allele in the outgroup, are
    ignored.

    This function is an implementation of the method mentioned in Baudry
    and Depaulis (2003), allowing to estimate the probability that a
    site oriented using the provided outgroup have be misoriented due to
    a homoplasic mutation in the branch leading to the outgroup. Note
    that this estimation neglects the probability of shared
    polymorphism.

    If the instance is created with an alignment as constructor argument,
    then the statistics are computed. The method :meth:`.load_align`
    does the same from an existing :class:`!ProbaMisoriented` instance.
    Otherwise, individual sites can be loaded with :meth:`.load_site`,
    and then the statistics can be computed using :meth:`.compute` (the
    latter is preferable if generate :class:`.Freq` instances for an
    other use).

    .. admonition:: Reference

        Baudry, E. & F. Depaulis. 2003. Effect of misoriented
        sites on neutrality tests with outgroup.
        *Genetics*\ \xa0\  **165**\ : 1619-1622.

    .. versionadded:: 3.0.0
    """

    def __init__(self, align=None, struct=None):
        self._A = alphabets.DNA.get_code('A')
        self._C = alphabets.DNA.get_code('C')
        self._G = alphabets.DNA.get_code('G')
        self._T = alphabets.DNA.get_code('T')
        self._sd = _eggwrapper.SiteDiversity()
        self._site = _site.Site()
        self._freq = _freq.Freq()
        self.reset()
        if align is not None: self.load_align(align, struct)

    def reset(self):
        """
        Clear all loaded or computed data.
        """
        self._S = 0
        self._D = 0
        self._Ti = 0
        self._Ti_cnt = 0
        self._TiTv = None
        self._pM = None
        self._M = 0
        self._M_cnt = 0

    def load_align(self, align, struct):
        """
        Load all sites of align that meet criteria. If there are previously
        loaded data, they are discarded. This method computes statistics.
        Data are required to be DNA sequences.

        :param align: an :class:`.Align` instance.
        :param struct: a :class:`.Structure` instance.
        """
        if align._alphabet._obj.get_type() != 'DNA': raise ValueError('data must be DNA sequences')
        self.reset()
        for i in range(align._obj.get_nsit_all()):
            self._site.from_align(align, i)
            self._freq.from_site(self._site, struct)
            self.load_site(self._freq)
        self.compute()

    def load_site(self, freq):
        """
        Load a single site. If there are previously loaded data, they
        are retained. To actually compute the misorientation
        probability, the user must call :meth:`.compute`.

        :param freq: a :class:`.Freq` instance.

        .. warning::
            The origin of data must be DNA sequences. This is currently
            not checked here.
        """
        flag = self._sd.process(freq._obj)
        if (flag&2) == 0: return
        if self._sd.Aglob() < 2: return # skip sites that are fixed overall
        if self._sd.Aing() > 2: return # skip sites with 3+ alleles ingroup
        if self._sd.Aout() > 1: return # skip sites with 2+ alleles outgroup
        if self._sd.Aing() == 2:
            self._S += 1
            if freq._obj.frq_outgroup().nseff() > 0:
                self._M_cnt += 1
                if self._sd.Aglob() > self._sd.Aing():
                    self._M += 1
        else:
            self._D += 1
        if self._sd.Aglob() == 2:
            self._Ti_cnt += 1
            if self._sd.global_allele(0) == self._A :
                if self._sd.global_allele(1) == self._G: self._Ti += 1
            elif self._sd.global_allele(0) == self._G:
                if self._sd.global_allele(1) == self._A:  self._Ti += 1
            elif self._sd.global_allele(0) == self._C:
                if self._sd.global_allele(1) == self._T: self._Ti += 1
            elif self._sd.global_allele(0) == self._T:
                if self._sd.global_allele(1) == self._C:  self._Ti += 1

    def compute(self):
        """
        Compute :obj:`.pM` and :obj:`.TiTv` statistics. Requires that
        sites have been loaded using :meth:`.load_site`. This method
        does not reset the instance.
        """
        self._pM = None
        self._TiTv = None
        if self._Ti_cnt > 0 and self._M_cnt > 0:
            Ti = self._Ti / self._Ti_cnt
            a = Ti / 4
            b = (1-Ti) / 8
            if b > 0:
                self._TiTv = a/b
                self._pM = ((self._M/self._M_cnt) * a**2 + 2*b**2) / (2*b*(2*a+b))

    @property
    def S(self):
        """ Number of loaded polymorphic sites. Only the ingroup is
            considered to classify a site as polymorphic. """
        return self._S

    @property
    def D(self):
        """
        Number of loaded sites with a fixed difference to the outgroup.
        """
        return self._D

    @property
    def TiTv(self):
        """
        Transition and transversion rate ratio. ``None`` if
        the value cannot be computed (no loaded data or null
        transversion rate). Requires that :meth:`.compute` has been
        called.
        """
        return self._TiTv

    @property
    def pM(self):
        """
        Probability of misorientation. ``None`` if the value cannot be
        computed (no loaded data, no valid polymorphism, null
        transversion rate). Requires that :meth:`.compute` has been
        called.
        """
        return self._pM



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, egglib, sys, unittest, random, re, gc, time
import collections

class Simulator_test(unittest.TestCase):
    #test on the creation of an object of the class "Simulator"
    def test_Simulator_T(self):
        sim = egglib.coalesce.Simulator(num_pop=4)
        self.assertIsInstance(sim,egglib.coalesce._simulator.Simulator)

    def test_Simulator_E(self):
        with self.assertRaises(TypeError):
            sim =egglib.coalesce.Simulator()

    #Test on the method "params" of the class "Simulator"
    def test_params_T(self):
        sim = egglib.coalesce.Simulator(num_pop=4)
        self.assertIsInstance(sim.params, egglib.coalesce._param_helpers.ParamDict)

    #Test on the method "simul" of the class "Simulator"
    def test_simul_T(self):
        sim = egglib.coalesce.Simulator(2, migr=0.1, theta=2.0, num_chrom=(40,40))
        aln=sim.simul()
        self.assertIsInstance(aln, egglib._interface.Align)
    
    def test_simul_E(self):
        sim = egglib.coalesce.Simulator(2)
        with self.assertRaises(ValueError):
            aln=sim.simul()

    #Test on the method "iter_simul" of the class "Simulator"
    def test_iter_simul_T(self):
        sim = egglib.coalesce.Simulator(2, migr=0.1, theta=2.0, num_chrom=(40,40))
        self.assertIsInstance(sim.iter_simul(10000), collections.abc.Iterable)


    def test_iter_simul_T2(self):
        egglib.random.set_seed(304059594)
        sim = egglib.coalesce.Simulator(2, migr=0.1, theta=2.0, num_chrom=(40,40))
        pi=[4.141455696202532,17.020886075949363,10.81455696202532,12.924683544303793,33.9405063291139,20.88607594936709,16.693354430379742,3.620569620253164,30.46993670886073,15.046835443037972]
        thetaw=[4.845568424245371,9.287339479803627,5.8550618459631565,6.8645552676809425,15.949996063141013,11.104427638895642,8.681643426772956,3.634176318184028,14.334806588392556,7.470251320711614]
        d=[-0.4431286862903224,2.6965022453054894,2.63744693521985,2.7910450471752473,3.7636271400744246,2.8855818820313806,2.9739780411465353,-0.010992404265106911,3.737735546441633,3.230129104396626]
        cs = egglib.stats.ComputeStats()
        cs.configure()
        cs.add_stats('Pi')
        cs.add_stats('thetaW')
        cs.add_stats('D')
        for items in sim.iter_simul(10, cs=cs):
            self.assertIn(items['Pi'], pi)
            self.assertIn(items['thetaW'], thetaw)
            self.assertIn(items['D'], d)

    #test on the creation of an object of the class "Align" with the class "Simulator"
    def test_sim_aln_T(self):
        coal = egglib.coalesce.Simulator(1, num_chrom=(20,), migr=0, theta=5.0)
        aln = coal.align
        self.assertIsInstance(aln, egglib.Align)

    def check(self, aln, ns, ls):
        self.assertEqual(aln.ns, ns)
        self.assertEqual(aln.ls, ls)

    def test_align_num_samples_outgroup(self):
        coal = egglib.coalesce.Simulator(1, num_chrom=(20,), num_mut=10)
        self.check(coal.simul(), 20, 10)
        self.check(coal.align, 20, 10)
        for aln in coal.iter_simul(1):
            self.check(aln, 20, 10)
        cs = egglib.stats.ComputeStats()
        cs.add_stats('nseff', 'lseff')
        for stats in coal.iter_simul(1, cs=cs):
            self.assertEqual(stats['nseff'], 20)
            self.assertEqual(stats['lseff'], 10)
            self.check(coal.align, 20, 10)

        coal.params.add_event('sample', 0.1, idx=0, num_chrom=20, num_indiv=0, label='1')
        coal.params.add_event('sample', 0.2, idx=0, num_chrom=12, num_indiv=0, label='64')
        self.check(coal.simul(), 52, 10)
        self.check(coal.align, 52, 10)
        for aln in coal.iter_simul(1):
            self.check(aln, 52, 10)
            self.check(coal.align, 52, 10)
        for stats in coal.iter_simul(1, cs=cs):
            self.assertEqual(stats['nseff'], 52)
            self.assertEqual(stats['lseff'], 10)
            self.check(coal.align, 52, 10)

        coal = egglib.coalesce.Simulator(2, num_chrom=(20,10), migr=1, num_mut=7)
        self.check(coal.simul(), 30, 7)
        self.check(coal.align, 30, 7)

    def test_alphabets(self): # just executing some code, not actually testing anything
        coal = egglib.coalesce.Simulator(1, theta=2, num_chrom=[20])
        aln = coal.simul()
        aln.alphabet.name
        for i in aln:
            i.sequence[:]
        aln.fasta(alphabet=egglib.alphabets.DNA)

        coal.params['mut_model'] = 'IAM'
        aln = coal.simul()
        aln.alphabet.name
        for i in aln:
            i.sequence[:]

        coal.params['mut_model'] = 'TPM'
        aln = coal.simul()
        aln.alphabet.name
        for i in aln:
            i.sequence[:]



"""
    Copyright 2015-2023 Stephane De Mita, Mathieu Siol, Thomas Coudoux

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from .. import eggwrapper as _eggwrapper
from .. import alphabets
from .. import _site

FIRST = _eggwrapper.FIRST
LAST = _eggwrapper.LAST

### NOTE: it is necessary to replace the position() method of _eggwrapper.VcfParser!!! 

class _VcfParserBase(object):
    def _f_position(self):
        x = _eggwrapper.VcfParser.position(self._parser)
        return -1 if x == _eggwrapper.BEFORE else x

    def _position(self):
        raise NotImplementedError

    def __init__(self):
        raise NotImplementedError

    @staticmethod
    def _mk_index_fname(fname):
        basename, ext = os.path.splitext(fname)
        if ext == '.vcf': return basename + '.vcfi'
        else: return fname + '.vcfi'

    def _set_params(self, threshold_PL, threshold_GL):
        if threshold_PL is None and threshold_GL is None:
            self._parser.set_threshold_PL(_eggwrapper.UNKNOWN)
            self._parser.set_threshold_GL(_eggwrapper.UNKNOWN)
        elif threshold_PL is not None and threshold_GL is None :
            if threshold_PL < 1: raise ValueError('threshold_PL must be at least 1')
            else: self._parser.set_threshold_PL(threshold_PL) 
        elif threshold_GL is not None and threshold_PL is None :
            if threshold_GL < 1: raise ValueError('threshold_GL must be at least 1')
            else: self._parser.set_threshold_GL(threshold_GL)            
        else:
            raise ValueError('cannot use threshold_PL and threshold_GL at the same time')

    file_format = property(lambda self: self._parser.file_format(), doc='File format specified in the header.')
    num_info = property(lambda self: self._parser.num_info(), doc='Number of INFO definitions in the header.')
    num_format = property(lambda self: self._parser.num_format(), doc='Number of FORMAT definitions in the header.')
    num_filter = property(lambda self: self._parser.num_filter(), doc='Number of FILTER definitions in the header.')
    num_alt = property(lambda self: self._parser.num_alt(), doc='Number of ALT definitions in the header.')
    num_meta = property(lambda self: self._parser.num_meta(), doc='Number of META definitions in the header.')
    num_samples = property(lambda self: self._parser.num_samples(), doc='Number of samples defined in the header.')

    def get_sample(self, idx):
        """
        Get the name of a sample read from the header. The index
        must be smaller than :py:obj:`~.io.VcfParser.num_samples`.
        """
        if idx < 0 or idx >= self._parser.num_samples(): raise IndexError('sample index out of range')
        return self._parser.get_sample(idx)

    def get_info(self, idx):
        """
        Get an INFO definition from the header. The
        index must be smaller than :py:obj:`.num_info`. Return a
        dictionary containing the following data:

        * ``"id"``: identifier string.
        * ``"type"``: one of ``"Integer"``, ``"Float"``, ``"Flag"``,
          ``"Character"``, and ``"String"``.
        * ``"description"``: description string.
        * ``"number"``: expected number of items. Special values are
          ``None`` (if undefined), ``"NUM_GENOTYPES"`` (number matching
          the number of genotypes for any particular variant),
          ``"NUM_ALTERNATE"`` (number matching the number of alternate
          alleles for any particular variant), and ``"NUM_ALLELES"``
          (number matching the number of alleles--including the
          reference--for any particular variant).
        * ``"extra"``: all extra qualifiers, presented as a list
          of ``(key, value)`` :class:`tuple` instances.
        """
        if idx < 0 or idx >= self._parser.num_info(): raise IndexError('INFO index out of range')
        info = self._parser.get_info(idx)
        n = info.get_number()
        if n == _eggwrapper.UNKNOWN: n = None
        elif n == _eggwrapper.NUM_ALTERNATE: n = 'NUM_ALTERNATE'
        elif n == _eggwrapper.NUM_GENOTYPES: n = 'NUM_GENOTYPES'
        elif n == _eggwrapper.NUM_POSSIBLE_ALLELES: n = 'NUM_ALLELES'
        t = info.get_type()
        if t == _eggwrapper.Info.Integer: t = 'Integer'
        elif t == _eggwrapper.Info.Float: t = 'Float'
        elif t == _eggwrapper.Info.Flag: t = 'Flag'
        elif t == _eggwrapper.Info.Character: t = 'Character'
        elif t == _eggwrapper.Info.String: t = 'String'
        else: raise ValueError('invalid VCF metainformation type found')
        return {
            'id': info.get_ID(),
            'description': info.get_description(),
            'number': n,
            'extra': [(info.get_extra_key(j), info.get_extra_value(j))
                                for j in range(info.get_num_extra())],
            'type': t
        }

    def get_format(self, idx):
        """
        Get a FORMAT definition from the header. The
        index must be smaller than :py:obj:`~.io.VcfParser.num_format`. Return a
        dictionary containing the following data:

        * ``"id"``: identifier string.
        * ``"type"``: one of ``"Integer"``, ``"Float"``, ``"Character"``,
          and ``"String"``.
        * ``"description"``: description string.
        * ``"number"``: expected number of items. Special values are
          ``None`` (if undefined), ``"NUM_GENOTYPES"`` (number matching
          the number of genotypes for any particular variant),
          ``"NUM_ALTERNATE"`` (number matching the number of alternate
          alleles for any particular variant), and ``"NUM_ALLELES"``
          (number matching the number of alleles--including the
          reference--for any particular variant).
        * ``"extra"``: all extra qualifiers, presented as a list
          of ``(key, value)`` :class:`tuple` instances.
        """

        if idx < 0 or idx >= self._parser.num_format(): raise IndexError('FORMAT index out of range')
        format_ = self._parser.get_format(idx)

        n = format_.get_number()
        if n == _eggwrapper.UNKNOWN: n = None
        elif n == _eggwrapper.NUM_ALTERNATE: n = 'NUM_ALTERNATE'
        elif n == _eggwrapper.NUM_GENOTYPES: n = 'NUM_GENOTYPES'
        elif n == _eggwrapper.NUM_POSSIBLE_ALLELES: n = 'NUM_ALLELES'

        t = format_.get_type()
        if t == _eggwrapper.Info.Integer: t = 'Integer'
        elif t == _eggwrapper.Info.Float: t = 'Float'
        elif t == _eggwrapper.Info.Character: t = 'Character'
        elif t == _eggwrapper.Info.String: t = 'String'
        else: raise ValueError('invalid VCF metainformation type found')

        return {
            'id': format_.get_ID(),
            'description': format_.get_description(),
            'number': n,
            'extra': [(format_.get_extra_key(j), format_.get_extra_value(j))
                                for j in range(format_.get_num_extra())],
            'type': t
        }

    def get_filter(self, idx):
        """
        Get a FILTER definition from the header. The
        index must be smaller than :py:obj:`~.io.VcfParser.num_filter`. Return a
        dictionary containing the following data:

        * ``"id"``: identifier string.
        * ``"description"``: description string.
        * ``"extra"``: all extra qualifiers, presented as a list
          of ``(key, value)`` :class:`tuple` instances.
        """
        if idx < 0 or idx >= self._parser.num_filter(): raise IndexError('FILTER index out of range')
        filter_ = self._parser.get_filter(idx)
        return {
            'id': filter_.get_ID(),
            'description': filter_.get_description(),
            'extra': [(filter_.get_extra_key(j), filter_.get_extra_value(j))
                                for j in range(filter_.get_num_extra())]
        }

    def get_alt(self, idx):
        """
        Get an ALT definition from the header. The
        index must be smaller than :py:obj:`.num_alt`. Return a
        dictionary containing the following data:

        * ``"id"``: identifier string.
        * ``"description"``: description string.
        * ``"extra"s``: all extra qualifiers, presented as list
          of ``(key, value)`` :class:`tuple` instances.
        """
        if idx < 0 or idx >= self._parser.num_alt(): raise IndexError('ALT index out of range')
        alt = self._parser.get_alt(idx)
        return {
            'id': alt.get_ID(),
            'description': alt.get_description(),
            'extra': [(alt.get_extra_key(j), alt.get_extra_value(j))
                                for j in range(alt.get_num_extra())]
        }

    def get_meta(self, idx):
        """
        Get data for a given META field defined in the VCF header. The
        index must be smaller than :py:obj:`~.io.VcfParser.num_meta`. Return a
        :class:`tuple` containing the key and the value of the META
        field.
        """
        if idx < 0 or idx >= self._parser.num_meta(): raise IndexError('META index out of range')
        meta = self._parser.get_meta(idx)
        return ( meta.get_key(), meta.get_value() )

    def get_variant(self):
        """
        Get all data for the current variant.
        Return an :class:`.io.VcfVariant` instance containing all data available
        for the last variant processed by this instance. It is required
        that a variant has been effectively processed.
        """
        return VcfVariant._make(self)

    def get_genotypes(self, start=0, stop=None, dest=None):
        """
        Extract genotype data for the current site.

        It is required that a variant has been effectively processed.

        :param start: index of the first sample to process (at least
            0 and smaller than the number of samples).

        :param stop: sample index at which processing must be
            stopped (this sample is not processed; at least equal
            to *start* and smaller than the number of samples). 

        :param dest: a :class:`.Site`
            instance that will be recycled and used to place results.

        :return: A :class:`.Site` instance by default, or
            ``None`` if *dest* was specified.
        """

        # check parser
        if not self._parser.has_data(): raise ValueError('data must have been read from VCF parser')
        if not self._parser.has_GT(): raise ValueError('`VcfParser` instance must have `GT` data')
        pl = self._parser.ploidy()
        if pl < 1: raise ValueError('GT ploidy is 0')

        # check indexes
        n = self._parser.num_samples()
        if start < 0 or start >= n: raise IndexError('invalid start index')
        if stop is None: stop = n
        elif stop < start or stop > n: raise IndexError('invalid stop index')

        # create a new site instance or recycle if needed
        if dest is None: site = _site.Site()
        else: site = dest
        siteobj = site._obj
        siteobj.reset()
        site._alphabet = self._mk_alphabet()
        site._position = self._parser.position()

        # process
        siteobj.process_vcf(self._parser, start, stop)

        # create/recycle provided instance
        if dest is None:
            return site

    def _mk_alphabet(self):
        if self._parser.type_alleles() == 0:
            return alphabets.DNA
        if self._parser.type_alleles() == 1:
            alph = alphabets.Alphabet('string', [], [], case_insensitive=True, name=None)
            self._parser.set_alleles(alph._obj)
            return alph
        if self._parser.type_alleles() == 2 or self._parser.type_alleles() == 3:
            alph = alphabets.Alphabet('custom', [], [], case_insensitive=False, name=None)
            self._parser.set_alleles(alph._obj)
            return alph
        raise RuntimeError('unexpected value for `type_alleles` (internal error)')

class VcfStringParser(_VcfParserBase):
    """
    Import Variant Call Format data from a string.
    Alias of :class:`.io.VcfParser` processing strings instead of a file.
    The constructor takes a string containing a VCF header (the first
    line being the file format specification and the last line being the
    header line ,starting with ``#CHROM``).
    """

    def __init__(self, header, threshold_PL=None, threshold_GL=None):
        string = header.strip()
        self._parser = _eggwrapper.VcfParser()
        self._position = self._f_position # substitution function (replace BEFORE by -1)
        self._parser.read_header(string)
        self._set_params(threshold_PL, threshold_GL)

    def readline(self, string):
        """
        Read one variant from a user-provided single line. The string
        should contain a single line of VCF-formatted data (no header).
        All field specifications and sample information should be
        consistent with the information contained in the header that
        has been provided when creating this instance.

        :return: A :class:`tuple` with the chromosome name, the position, and
          the number of alleles at the variant (as indicated in the VCF
          line).
        """
        self._parser.readline(string)
        return self._parser.chromosome(), self._position(), self._parser.num_alternate() + 1

class VcfParser(_VcfParserBase):
    """
    Import Variant Call Format data from a file. The VCF format
    is designed to store data describing genomic variation in an
    extremely flexible way.
    See the `description of the VCF format. <https://github.com/samtools/hts-specs>`_
    To parse VCF data stored in string, use :class:`.io.VcfStringParser`.

    :param fname: name of a properly formatted VCF file. The header
        section will be processed upon instance creation, and lines will
        be read later, when the user iterates over the instance (or call
        :meth:`~.io.VcfParser.readline`).

    :param threshold_PL: call genotypes parameter.
        This parameter controls how genotype calling (GT field) is
        performed from the PL (phred-scaled genotype likelihood) field.
        By default (``None``), this conversion is never performed. If
        *threshold_PL* is specified, genotype calling is
        only performed if GT is not available and PL is available. The
        genotype with the lowest PL value is selected. The parameter gives the
        minimum acceptable gap between the best genotype and the second one.
        If the second genotype has a too good score, based on this parameter,
        the genotype is called unknown. The parameter must be at least 1.

    :class:`.io.VcfParser` instances are iterable. Everly loop
    yields a ``(chromosome, position, num_all)``
    :class:`tuple` that allows the user to determines if the variant is of
    interest. Note that the position is considered as an index and therefore
    has been decremented compared with the value found in the file.
    If the variant is of interest, :class:`.io.VcfParser` instances provide methods to
    extract all data for this variant. Iterating over VCF lines can be
    performed manually using the method :meth:`~.io.VcfParser.readline`.
    """

    def __init__(self, fname, threshold_PL=None, threshold_GL=None):
        self._parser = _eggwrapper.VcfParser()
        self._position = self._f_position # substitution function (replace BEFORE by -1)
        if not isinstance(fname, str): raise TypeError('invalid fname: invalid type')
        self._parser.open_file(fname)
        self._fname = fname
        self._set_params(threshold_PL, threshold_GL)

    def close(self):
        """
        Close file (if it is open)
        """
        self._parser.reset()

    @property
    def good(self):
        """
        Tell if the file is good for reading. ``False`` is the end of
        the file has been reached.
        """
        return self._parser.good()

    @property
    def currline(self):
        """ Index of the current line of the VCF file. """
        return self._parser.get_currline()

    def __iter__(self):
        return self

    def __next__(self):
        if not self._parser.good(): raise StopIteration
        self._parser.read()
        return self._parser.chromosome(), self._position(), self._parser.num_alternate() + 1

    def readline(self):
        """
        Read one variant.
        Return a :class:`tuple` with the chromosome name, the position, and
        the number of alleles at the variant (as indicated in the VCF
        file).
        Raise a :exc:`ValueError` if file is finished.
        """
        if not self._parser.good(): raise ValueError('cannot read line')
        self._parser.read()
        return self._parser.chromosome(), self._position(), self._parser.num_alternate() + 1

    def load_index(self, fname=None):
        """
        Load an index file allowing fast navigation.
        Index files allow fast navigation in VCF file regardless of
        their size, allowing to move instantly to a determined variant
        using its position or its index in file. Index files can be
        created by :func:`.io.make_vcf_index`.

        :param fname: name of the index file. By default, use the default
            index file name, which is the name of the VCF file with the
            ".vcfi" extension (removing the original extension only if it
            is ".vcf").
        """
        if fname is None:
            fname = self._mk_index_fname(self._fname)
        if not os.path.exists(fname):
            raise IOError('index file does not exist')
        self._parser.get_index().load_data(self._parser, fname)

    @property
    def num_index(self):
        """
        Number of indexed lines if the file index.
        """
        return self._parser.get_index().num()

    @property 
    def has_index(self):
        """
        ``True`` if an index file has been loaded.
        """
        return self._parser.get_index().has_data()

    def goto(self, contig, position=FIRST):
        """goto(self, contig, position=egglib.io.FIRST)

        Move to an arbitrary position of the VCF file.
        Requires an index (see :meth:`~.io.VcfParser.load_index`).
        A :exc:`ValueError` is raised if the position cannot be found.

        :param contig: contig name.
        :param position: contig position. Use :py:obj:`.io.FIRST` for the
            first available variant of the contig, :py:obj:`.io.LAST` for
            the last, and -1 for the position immediately before the
            first position (position 0 in VCF files).
        """
        self._parser.get_index().go_to(contig, _eggwrapper.BEFORE if position == -1 else position)

    def unread(self):
        """
        Go back to the previous variant. No index is required, but only
        allowed after reading one line (not allowed immediately after
        creating instance or calling :meth:`~.io.VcfParser.rewind`).
        """
        self._parser.unread()

    def rewind(self):
        """
        Reset the parser. Move back to the first variant.
        No index is required.
        """
        self._parser.rewind()

    def slider(self, size, step, as_variants=False, start=0, stop=None, max_missing=0, allow_indel=False, allow_custom=False):
        """
        Start a sliding window from the current position.

        :param size: size of the sliding window (by default, in base pairs).
        :param step: increment of the sliding window (by default, in base pairs).
        :param as_variants: express size and step in number of variants instead of base pairs.
        :param start: start position of the sliding window.
        :param stop: stop position of the sliding window.
        :param max_missing: maximum number of missing alleles (variants
            exceeding this threshold are ignored).

            .. warning::
                Here, *max_missing* must be an absolute number of
                samples.

        :param allow_indel: include variants with alleles of variable size
        :param allow_custom: include variants with custom alleles

        :return: An :class:`.io.VcfSlidingWindow` instance.

        .. versionchanged:: 3.2.0
            *max_missing* is required to be an integer.
        """
        if not self._parser.good(): raise ValueError('parser reached end of file')
        if max_missing > self._parser.num_samples(): raise ValueError('invalid value for `max_missing`')
        if not isinstance(max_missing, int): raise TypeError('`max_missing\' must be an integer')
        obj = VcfSlidingWindow.__new__(VcfSlidingWindow)
        obj._sld = _eggwrapper.VcfWindowSliderPerSite() if as_variants else _eggwrapper.VcfWindowSlider()
        obj._wdw = VcfWindow.__new__(VcfWindow)
        obj._wdw._sld = obj._sld
        obj._wdw._parser = self
        obj._parser = self._parser
        if stop is None: stop = _eggwrapper.UNKNOWN
        if start < 0: raise ValueError('invalid start value')
        if stop < start: raise ValueError('invalid stop value')
        ns = self._parser.num_samples()
        mask = 3
        if allow_indel: mask &= 2
        if allow_custom: mask &= 1
        obj._sld.setup(self._parser, size, step, start, stop, max_missing, mask)
        return obj

    def bed_slider(self, bed, max_missing=0, allow_indel=False, allow_custom=False):
        """
        Perform a sliding window based on BED coordinates.

        :param bed: an :class:`.io.BED` instance.
        :param max_missing: maximum number of missing alleles.

            .. warning::
                Here, *max_missing* must be an absolute number of
                samples.
        :param allow_indel: include variants with alleles of varying size.
        :param allow_custom: include variants with custom alleles.

        :return: An :class:`.io.VcfWindow` instance.

        .. versionchanged:: 3.2.0
            *max_missing* is required to be an integer.
        """
        if not self._parser.good(): raise ValueError('parser reached end of file')
        if not isinstance(max_missing, int): raise TypeError('`max_missing\' must be an integer')
        if max_missing > self._parser.num_samples(): raise ValueError('invalid value for `max_missing`')
        obj = VcfSlidingWindow.__new__(VcfSlidingWindow)
        obj._sld = _eggwrapper.VcfWindowBED()
        obj._wdw = VcfWindow.__new__(VcfWindow)
        obj._wdw._sld = obj._sld
        obj._wdw._parser = self
        obj._parser = self._parser
        ns = self._parser.num_samples()
        mask = 3
        if allow_indel: mask &= 2
        if allow_custom: mask &= 1
        obj._sld.setup(self._parser, bed._obj, max_missing, mask)
        obj._bed_reference = bed # keep a reference to BED slider to avoid intempestive garbage collection
        return obj

def make_vcf_index(fname, outname=None):
    """
    Create the index for a VCF file. 

    :param fname: name of a VCF file.
    :param outname: name of the index file. By default, use a default
        name. The default name is based on the VCF file name, stripping
        the extension only if it is vcf and appending the vcfi
        extension. Index files bearing this default name are
        automatically loaded if the corresponding VCF file is opened.
    """
    _parser = _eggwrapper.VcfParser()
    _parser.open_file(fname)
    if outname is None:
        outname = VcfParser._mk_index_fname(fname)
    _eggwrapper.make_vcf_index(_parser, outname)
    del _parser
    return outname

class VcfVariant(object):
    """
    Represent a single VCF variant.
    The user cannot create instances of this class himself
    (instances are generated by :class:`.io.VcfParser` or :class:`.io.VcfStringParser`).

    .. note::

        The ``AA`` (ancestral allele), ``AN`` (allele number),
        ``AC`` (allele count), and ``AF`` (allele frequency) INFO fields
        as well as the ``GT`` (deduced genotype) FORMAT are
        automatically extracted if they are present in the the file and
        if their definition matches the format specification (meaning
        that they were not re-defined with different number/type) in
        the header. If present, they are available through the dedicated
        attributes :py:obj:`~.io.VcfVariant.AN`, :py:obj:`~.io.VcfVariant.AA`,
        :py:obj:`~.io.VcfVariant.AC`, :py:obj:`~.io.VcfVariant.AF`,
        :py:obj:`~.io.VcfVariant.GT`, :py:obj:`~.io.VcfVariant.ploidy`
        and :py:obj:`~.io.VcfVariant.GT_phased`. However,
        they are also available in the respective
        :py:obj:`~.io.VcfVariant.info` and
        :py:obj:`~.io.VcfVariant.samples` (sub)-dictionaries.
    """

    # the following are provided for comparison to content of variant.alternate_types
    alt_type_default = _eggwrapper.Default #: Explicit alternate allele (the string represents the nucleotide sequence of the allele).
    alt_type_referred = _eggwrapper.Referred #: Alternate allele referring to a pre-defined allele (the string provides the ID of the allele).
    alt_type_breakend = _eggwrapper.Breakend #: Alternate allele symbolizing a breakend (see VCF description for more details).

    def __init__(self):
        raise NotImplementedError('cannot create `VcfVariant` instance')

    @staticmethod
    def _filter_missing(v, type_):
        if ((type_==int and v==_eggwrapper.MISSINGDATA) or
            (type_==float and v==_eggwrapper.UNDEF) or
            (type_==str and v[0]==_eggwrapper.MAXCHAR)): return None
        else: return v

    @classmethod
    def _make(cls, parser):
        if parser._parser.len_reference() == 0: raise ValueError('cannot generate `VcfVariant` instance: no VCF line has been parsed (or the length of the reference allele is null)')
        obj = cls.__new__(cls)

        # get chromosome
        obj._chrom = parser._parser.chromosome()
        if obj._chrom == '': obj._chrom = None

        # get position
        obj._pos = parser._position()
        if obj._pos == _eggwrapper.MISSING: obj._pos = None
        elif obj._pos == _eggwrapper.UNKNOWN: obj._pos = -1

        # get IDs
        obj._id = tuple(parser._parser.ID(i) for i in range(parser._parser.num_ID()))

        # get reference+alternate alleles
        obj._alleles = [parser._parser.reference()]
        if obj._alleles[0] == '': obj._alleles[0] = None
        obj._alternate_types = []
        obj._num_alternate = parser._parser.num_alternate()
        obj._num_alleles = obj._num_alternate + 1
        for i in range(obj._num_alternate):
            obj._alleles.append(parser._parser.alternate(i))
            obj._alternate_types.append(parser._parser.alternate_type(i))
        obj._alleles = tuple(obj._alleles)
        obj._alternate_types = tuple(obj._alternate_types)

        # get quality
        obj._quality = parser._parser.quality()
        if obj._quality == _eggwrapper.UNDEF: obj._quality = None

        # get test info
        n = parser._parser.num_failed_tests()
        if n == _eggwrapper.UNKNOWN: obj._failed_tests = None
        else: obj._failed_tests = tuple(parser._parser.failed_test(i) for i in range(n))

        # get info for the whole variant (site)
        obj._info = {}
        for i in range(parser._parser.num_FlagInfo()):
            item = parser._parser.FlagInfo(i)
            obj._info[item.get_ID()] = ()
        for i in range(parser._parser.num_IntegerInfo()):
            item = parser._parser.IntegerInfo(i)
            if item.get_expected_number() == 1:
                obj._info[item.get_ID()] = cls._filter_missing(item.item(0), int)
            else:
                obj._info[item.get_ID()] = tuple(cls._filter_missing(item.item(j), int) for j in range(item.num_items()))

        for i in range(parser._parser.num_FloatInfo()):
            item = parser._parser.FloatInfo(i)
            if item.get_expected_number() == 1:
                obj._info[item.get_ID()] = cls._filter_missing(item.item(0), float)
            else:
                obj._info[item.get_ID()] = tuple(cls._filter_missing(item.item(j), float) for j in range(item.num_items()))

        for i in range(parser._parser.num_CharacterInfo()):
            item = parser._parser.CharacterInfo(i)
            if item.get_expected_number() == 1:
                obj._info[item.get_ID()] = cls._filter_missing(item.item(0), str)
            else:
                obj._info[item.get_ID()] = tuple(cls._filter_missing(item.item(j), str) for j in range(item.num_items()))
        for i in range(parser._parser.num_StringInfo()):
            item = parser._parser.StringInfo(i)
            if item.get_expected_number() == 1:
                obj._info[item.get_ID()] = cls._filter_missing(item.item(0), str)
            else:
                obj._info[item.get_ID()] = tuple(cls._filter_missing(item.item(j), str) for j in range(item.num_items()))

        # get predefined AN/AA/AC/AF info fields if they are matching definition
        if parser._parser.has_AN(): obj._AN = parser._parser.AN()
        else: obj._AN = None
        if parser._parser.has_AA(): obj._AA = parser._parser.AA_string()
        else: obj._AA = None
        if parser._parser.has_AC(): obj._AC = tuple(parser._parser.AC(i) for i in range(parser._parser.num_AC()))
        else: obj._AC = None
        if parser._parser.has_AF(): obj._AF = tuple(parser._parser.AF(i) for i in range(parser._parser.num_AF()))
        else: obj._AF = None

        # pre-process the format column to allow match the ID's to the proper accessor method of SampleInfo
        fields = []
        for i in range(parser._parser.num_fields()):
            format_ = parser._parser.field(i)
            type_ = format_.get_type()
            if type_ == _eggwrapper.Info.String:
                f1 = _eggwrapper.SampleInfo.num_StringItems
                f2 = _eggwrapper.SampleInfo.StringItem
                t = str
            elif type_ == _eggwrapper.Info.Float:
                f1 = _eggwrapper.SampleInfo.num_FloatItems
                f2 = _eggwrapper.SampleInfo.FloatItem
                t = float
            elif type_ == _eggwrapper.Info.Integer:
                f1 = _eggwrapper.SampleInfo.num_IntegerItems
                f2 = _eggwrapper.SampleInfo.IntegerItem
                t = int
            elif type_ == _eggwrapper.Info.Character:
                f1 = _eggwrapper.SampleInfo.num_CharacterItems
                f2 = _eggwrapper.SampleInfo.CharacterItem
                t = str
            fields.append((
                format_.get_ID(), # the ID
                parser._parser.field_rank(i), # the index within range
                f1, # SampleInfo's method to get number of items
                f2, # SampleInfo's method to get a given item
                t   # type (required to convert missing data to None)
            ))
        obj._format_fields = frozenset(i[0] for i in fields)

        # get all SampleInfo data
        obj._num_samples = parser._parser.num_samples()
        obj._samples = []
        for sam in range(obj._num_samples):
            sample_info = parser._parser.sample_info(sam)
            sample_data = {}
            for ID, idx, f1, f2, t in fields:
                sample_data[ID] = (
                    None if f1(sample_info, idx)==0 else
                    tuple(cls._filter_missing(f2(sample_info, idx, i), t) for i in range(f1(sample_info, idx))))
                    # the above line get all items for a given FORMAT (using method pointers)
            obj._samples.append(sample_data)

        # get ploidy/num genotpes
        obj._ploidy = parser._parser.ploidy()
        obj._num_genotypes = parser._parser.num_genotypes()

        # get genotypes
        if parser._parser.has_GT():
            obj._gt = tuple(
                tuple((None if parser._parser.GT(i,j) == _eggwrapper.UNKNOWN
                            else obj._alleles[parser._parser.GT(i,j)])
                                    for j in range(obj._ploidy))
                                    for i in range(obj._num_samples))
            obj._gt_phased = tuple(parser._parser.GT_phased(i) for i in range(obj._num_samples))

            obj._gt_field = []
            gt_f = None
            for i in range(obj._num_samples):
                aa_f = []
                for j in range(obj._ploidy):
                    if parser._parser.GT(i,j) == _eggwrapper.UNKNOWN:
                        aa_f = None
                    else:
                        aa_f.append(parser._parser.GT(i,j))
                if aa_f is not None:
                    aa_fs = tuple(map(str, aa_f))
                    if not parser._parser.GT_phased(i):
                        gt_f = '/'.join(aa_fs)
                    else : gt_f = '|'.join(aa_fs)
                else:
                    gt_f = '.'
                obj._gt_field.append(gt_f)
            obj._gt_field= tuple(obj._gt_field)

        else:
            obj._gt = None
            obj._gt_phased = None

        # get PL
        if parser._parser.has_PL():
            obj._pl = tuple(
                    tuple(parser._parser.PL(i, j) for j in range(obj._num_genotypes))
                            for i in range(obj._num_samples))
        else:
            obj._pl = None

        # get GL
        if parser._parser.has_GL():
            obj._gl = tuple(
                    tuple(parser._parser.GL(i, j) for j in range(obj._num_genotypes))
                            for i in range(obj._num_samples))
        else:
            obj._gl = None


        return obj

    # block of accessors below: they don't DO anything (it's only doc who makes is cumbersome)
    chromosome = property(lambda self: self._chrom, doc='Chromosome name. ``None`` if missing.')
    position = property(lambda self: self._pos, doc='Position. Given as an index; first value is 0. ``None`` if missing.')
    ID = property(lambda self: self._id, doc='Tuple containing all IDs.')
    num_alleles = property(lambda self: len(self._alleles), doc='Number of alleles. The reference is always included.')
    num_alternate = property(lambda self: self._num_alternate, doc='Number of alternate alleles. Equal to :py:obj:`~.io.VcfVariant.num_alleles` minus 1.')
    alleles = property(lambda self: self._alleles, doc='Tuple of variant alleles. The first is the reference, which is not guaranteed to be present in samples.')
    alternate_types = property(lambda self: self._alternate_types, doc="""Alternate alleles types, as a tuple.
One value is provided for each alternate allele. The provided values are
integers whose values should always be compared to class attributes
:py:obj:`~.io.VcfVariant.alt_type_default`, :py:obj:`~.io.VcfVariant.alt_type_referred` and
:py:obj:`~.io.VcfVariant.alt_type_breakend`, as in (for the type of the first alternate
allele)::

    >>> type_ = variant.alternate_types[0]
    >>> if type_ == variant.alt_type_default:
    ...     allele = variant.allele(0)""")
    quality = property(lambda self: self._quality, doc='Variant quality. ``None`` if missing.')
    failed_tests = property(lambda self: self._failed_tests, doc='Filters at which this variant failed. Provided as a tuple or names, or ``None`` if no filters have been applied.')
    AA = property(lambda self: self._AA, doc='Value of the AA info field. ``None`` if missing.')
    AN = property(lambda self: self._AN, doc='Value of the AN info field. ``None`` if missing.')
    AC = property(lambda self: self._AC, doc='Value of the AC info field. Provided as a tuple. ``None`` if missing.')
    AF = property(lambda self: self._AF, doc='Value of the AF info field. Provided as a tuple. ``None`` if missing.')
    info = property(lambda self: self._info, doc="""Dictionary of INFO
fields for this variant. Keys are ID of INFO fields available for this
variant, and values are always a :class:`tuple` of items, even if there is only
one item. For flag INFO types,
the value is always an empty :class:`!tuple`.""")
    format_fields = property(lambda self: self._format_fields, doc='Frozenset of FORMAT fields ID\'s. The frozenset is empty if no sample data are available.')
    num_samples = property(lambda self: self._num_samples, doc='Number of samples. Equivalent to ``len(variant.samples)``.')
    samples = property(lambda self: self._samples, doc="""Information
for each sample. Empty list if no samples are
defined. Otherwise, the list contains one dictionary for each sample: keys of
these dictionaries are FORMAT fields identifiers (the keys are always the same as
the content of :py:obj:`~.io.VcfVariant.format_fields`), and their values are :class:`tuple` instances in
all cases.""")
    ploidy = property(lambda self: self._ploidy, doc='Ploidy among genotypes. Always 2 if GT is not available.')
    num_genotypes = property(lambda self: self._num_genotypes, doc='Number of genotypes.')
    GT_phased = property(lambda self: self._gt_phased, doc='Tell if the genotype for each sample is phased. ``None`` if GT is not available.')
    GT = property(lambda self: self._gt, doc="""Genotypes from GT
fields. Only if this format field is available. Provided as a :class:`tuple` of
sub-tuples. The number of sub-tuples is equal to the number of samples
(:py:obj:`~.io.VcfVariant.num_samples`). The number of items within each
sub-tuples is equal to the ploidy (:py:obj:`~.io.VcfVariant.ploidy`). These items are
allele expression (as found in :py:obj:`~.io.VcfVariant.alleles`), or ``None`` (for
missing values). This attribute is ``None`` if GT is not available.""")

    GT_vcf = property(lambda self: self._gt_field, doc = """GT field as written in the VCF file""")
    PL = property(lambda self: self._pl, doc = """PL values for all samples""")
    GL = property(lambda self: self._gl, doc = """GL values for all samples""")

class VcfSlidingWindow(object):
    """
    This class manages a sliding window on a VCF file.
    This class cannot be instanciated directly: instances are
    returned by the methods :meth:`~.io.VcfParser.slider` and
    :meth:`~.io.VcfParser.bed_slider` of :class:`.io.VcfParser`
    instances.

    :class:`.io.VcfSlidingWindow` instances are iterable:
    iteration steps return a common :class:`.io.VcfWindow` instance
    which is updated at each iteration round.
    """

    def __init__(self):
        raise NotImplementedError('cannot create `VcfSlidingWindow` instance')

    @property
    def good(self):
        """
        ``True`` if there is still data to be processed.
        """
        return self._sld.good()

    def __iter__(self):
        return self

    def __next__(self):
        if self._sld.good():
            self._sld.next_window()
            return self._wdw
        else: raise StopIteration

    next = __next__ # for python2

class VcfWindow(object):
    """
    Provide access to sites of a window from a VCF file. The following
    operations are supported by :class:`.io.VcfWindow` instances:

    +-----------------------------+------------------------------------------------+
    | Operation                   | Result                                         |
    +=============================+================================================+
    | ``len(win)``                | number of sites in the window                  |
    +-----------------------------+------------------------------------------------+
    | ``win[i]``                  | get a site as a :class:`.Site` instance        |
    +-----------------------------+------------------------------------------------+
    | ``for site in win:``        | iterate over sites as :class:`.Site` instances |
    +-----------------------------+------------------------------------------------+
    """

    def __init__(self):
        raise NotImplementedError('cannot create `VcfWindow` instance')

    num_sites = property(lambda self: self._sld.num_sites(), doc='Number of sites in window.')
    chromosome = property(lambda self: self._sld.chromosome(), doc='Name of the chromosome or contig.')

    @property
    def bounds(self):
        """Bounds of the window. The second bound is not included in the window."""
        return self._sld.win_start(), self._sld.win_stop()

    def __len__(self):
        return self._sld.num_sites()

    def __iter__(self):
        site = self._sld.first_site()
        while site is not None:
            yield _site.Site._from_site_holder(site.site(), alphabets.Alphabet._make(site.alphabet()) if site.alphabet() else alphabets.DNA)
            site = site.next()

    def __getitem__(self, idx):
        if idx < 0:
            idx = self._sld.num_sites() + idx
        if idx >= self._sld.num_sites() or idx < 0:
            raise ValueError('invalid site index')
        else:
            site = self._sld.get_site(idx)
            return _site.Site._from_site_holder(site.site(), alphabets.Alphabet._make(site.alphabet()) if site.alphabet() else alphabets.DNA)

class BED(object):
    """
    Class holding BED (Browser Extensible Data) data.

    :param fname: name of the BED-formatted input file. By default,
        create an empty instance.

    BED is a flexible format representing
    genomic annotation. In the current implementation, only the
    chromosome/scaffold, start, and end positions (the first three
    fields of each lines, which are the only ones to be required) are
    processed. If incorrectly formed fields are provided besides the
    first three fields, the behaviour is undefined (that is, it is not
    guaranteed that are exception is raised).

    Supported operations:

    +-------------------+----------------------------+-------+
    | Operation         | Result                     | Notes |
    +===================+============================+=======+
    | ``len(bed)``      | Number of annotation items |       |
    +-------------------+----------------------------+-------+
    | ``bed[i]``        | Get an annotation          | (1)   |
    +-------------------+----------------------------+-------+
    | ``for i in bed:`` | Iterate over annotations   | (1)   |
    +-------------------+----------------------------+-------+

    Notes:

    #. Annotation items are represented by dictionaries containing the
       following keys: ``"chrom"``, ``"start"``, and ``"end"``.
    """

    def __init__(self, fname=None):
        self._obj = _eggwrapper.BedParser()
        if fname is not None:
            self._obj.get_bed_file(fname)

    def __len__(self):
        return self._obj.n_bed_data()

    def __getitem__(self, i):
        if i < 0:
            i = self._obj.n_bed_data() + i
        if i < 0 or i >= self._obj.n_bed_data():
            raise IndexError('invalid BED index')
        return {'chrom': self._obj.get_chrom(i),
                'start': self._obj.get_start(i),
                'end': self._obj.get_end(i)}

    def extend(self, values):
        """
        Append items from an iterable. All items of *values*
        must be a sequence containing a
        value for *chrom*, *start*, and *end* (in that order).
        """
        for item in values:
            self.append(* item)

    def append(self, chrom, start, end):
        """
        Add an item at the end of the BED data.
        """
        self._obj.append(chrom, start, end)



"""
    Copyright 2009-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from ._phyml import phyml
from ._codeml import codeml
from ._muscle import muscle, muscle3, muscle5
from ._clustal import clustal
from ._utils import paths
from ._blast import makeblastdb, megablast, dc_megablast, blastn, blastn_short, \
                    blastp, blastp_short, blastp_fast, blastx, blastx_fast, \
                    tblastn, tblastn_fast, tblastx, \
                    BlastHit, BlastHsp, BlastOutput, BlastQueryHits
from ._nj import nj
paths.load() # must be after loading all application modules



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class muscle5_test(unittest.TestCase):
    def setUp(self):
        self.dna = egglib.io.from_fasta(str(path / 'cds.fas'), alphabet=egglib.alphabets.DNA)
        del self.dna[self.dna.find('Poplar_LG_VIII_pseudogene', index=True)]
        cds = egglib.Container.create(self.dna)
        cds.to_codons()
        self.prot = egglib.tools.translate(cds)

    def test_shortcut(self):
        egglib.wrappers.muscle(self.dna)
        egglib.wrappers.muscle5(self.dna)

    def test_source_type(self):
        # Container
        egglib.wrappers.muscle5(self.dna)

        # Align
        self.dna.equalize()
        aln = egglib.Align.create(self.dna)
        egglib.wrappers.muscle5(aln)

        # protein
        egglib.wrappers.muscle5(self.prot)

        # list
        with self.assertRaisesRegex(TypeError, '`source\' must be a Container'):
            egglib.wrappers.muscle5(list(self.dna))

        # custom alphabet
        seq = egglib.Container.create([(sam.name, sam.sequence) for sam in self.dna], alphabet=egglib.Alphabet('char', expl='ACGT', miss='?'))
        with self.assertRaisesRegex(ValueError, 'alphabet must be DNA or protein'):
            egglib.wrappers.muscle5(seq)

        # invalid alphabet
        test = egglib.io.from_genepop(str(path / 'P93.gpop'))
        with self.assertRaisesRegex(ValueError, 'alphabet must be DNA or protein'):
            egglib.wrappers.muscle5(test)

    def test_perm(self):
        egglib.wrappers.muscle5(self.dna, perm='none')
        egglib.wrappers.muscle5(self.dna, perm='abc')
        egglib.wrappers.muscle5(self.dna, perm='acb')
        egglib.wrappers.muscle5(self.dna, perm='bca')
        egglib.wrappers.muscle5(self.prot, perm='none')
        egglib.wrappers.muscle5(self.prot, perm='abc')
        egglib.wrappers.muscle5(self.prot, perm='acb')
        egglib.wrappers.muscle5(self.prot, perm='bca')
        with self.assertRaisesRegex(ValueError, 'invalid value for `perm\''):
            egglib.wrappers.muscle5(self.dna, perm='tronc')
        with self.assertRaisesRegex(ValueError, 'invalid value for `perm\''):
            egglib.wrappers.muscle5(self.dna, perm=None)

    def test_perturb(self):
        egglib.wrappers.muscle5(self.dna, perturb=0)
        egglib.wrappers.muscle5(self.prot, perturb=0)
        egglib.wrappers.muscle5(self.dna, perturb=1)
        egglib.wrappers.muscle5(self.prot, perturb=1)
        egglib.wrappers.muscle5(self.dna, perturb=107143)
        egglib.wrappers.muscle5(self.prot, perturb=107143)
        with self.assertRaisesRegex(TypeError, '`perturb\' must be an integer'):
            egglib.wrappers.muscle5(self.dna, perturb=0.0)
        with self.assertRaisesRegex(TypeError, '`perturb\' must be an integer'):
            egglib.wrappers.muscle5(self.dna, perturb=None)
        with self.assertRaisesRegex(TypeError, '`perturb\' must be an integer'):
            egglib.wrappers.muscle5(self.dna, perturb='0')
        with self.assertRaisesRegex(ValueError, 'invalid value for `perturb\''):
            egglib.wrappers.muscle5(self.dna, perturb=-1)

    def test_consiters_refineiters(self):
        egglib.wrappers.muscle5(self.dna, consiters=1, refineiters=1)
        egglib.wrappers.muscle5(self.dna, consiters=3, refineiters=10)
        egglib.wrappers.muscle5(self.prot, consiters=3, refineiters=10)
        egglib.wrappers.muscle5(self.dna, consiters=4, refineiters=2)
        with self.assertRaisesRegex(TypeError, '`consiters\' must be an integer'):
            egglib.wrappers.muscle5(self.dna, consiters='2', refineiters=100)
        with self.assertRaisesRegex(TypeError, '`refineiters\' must be an integer'):
            egglib.wrappers.muscle5(self.dna, consiters=2, refineiters=None)
        with self.assertRaisesRegex(TypeError, '`refineiters\' must be an integer'):
            egglib.wrappers.muscle5(self.dna, consiters=2, refineiters=100.0)
        with self.assertRaisesRegex(ValueError, '`consiters\' must be strictly positive'):
            egglib.wrappers.muscle5(self.dna, consiters=0, refineiters=100)
        with self.assertRaisesRegex(ValueError, '`consiters\' must be strictly positive'):
            egglib.wrappers.muscle5(self.dna, consiters=-2, refineiters=100)
        with self.assertRaisesRegex(ValueError, '`refineiters\' must be strictly positive'):
            egglib.wrappers.muscle5(self.dna, consiters=2, refineiters=0)
        with self.assertRaisesRegex(ValueError, '`refineiters\' must be strictly positive'):
            egglib.wrappers.muscle5(self.dna, consiters=2, refineiters=-10)

    def test_threads(self):
        egglib.wrappers.muscle5(self.dna, threads=None)
        egglib.wrappers.muscle5(self.dna, threads=1)
        egglib.wrappers.muscle5(self.dna, threads=4)
        egglib.wrappers.muscle5(self.dna, threads=100000)
        with self.assertRaisesRegex(TypeError, '`threads\' must be an integer'):
            egglib.wrappers.muscle5(self.dna, threads='four')
        with self.assertRaisesRegex(ValueError, '`threads\' must be strictly positive'):
            egglib.wrappers.muscle5(self.dna, threads=0)

    def test_return(self):
        for ref in self.dna, self.prot:
            aln = egglib.wrappers.muscle5(ref, threads=100)
            self.assertTrue(isinstance(aln, egglib.Align))
            self.assertEqual(aln.ns, ref.ns)
            num_gaps = 0
            for sam in ref:
                res = aln.find(sam.name)
                seq = res.sequence.string()
                num_gaps += seq.count('-')
                self.assertEqual(seq.replace('-', ''), sam.sequence.string())
            self.assertGreater(num_gaps, 0)

class muscle3_test(unittest.TestCase):
    def setUp(self):
        self.seq = egglib.io.from_fasta(str(path / 'cds.fas'), egglib.alphabets.DNA, labels=True)
        self.codons = egglib.io.from_fasta(str(path / 'cds.fas'), egglib.alphabets.DNA, labels=True)
        for i in self.codons:
            if i.ls%3!=0: i.sequence = i.sequence[:i.ls//3*3]
        self.codons.to_codons()
        self.prot = egglib.tools.translate(self.codons)

    def test_muscle_T(self):
        aln0=egglib.wrappers.muscle(self.seq, verbose=False)
        aln1=egglib.wrappers.muscle(self.seq, verbose=False)
        aln2=egglib.wrappers.muscle(self.seq, verbose=False, maxiters=1, diags=True)
        aln3=egglib.wrappers.muscle(self.seq, maxiters=1, diags=True, seqtype='dna',
                        aa_profile='sv', distance1='kbit20_3')
        aln4=egglib.wrappers.muscle(self.prot, seqtype='protein', aa_profile='sv')
        aln5=egglib.wrappers.muscle(self.prot, seqtype='protein', aa_profile='le',
                    brenner=True, diags1=True, SUEFF=0.4, anchorspacing=5,
                    center=-1.9, diaglength=17, distance1='kmer20_4',
                    gapopen=-2.5, maxtrees=2, objscore='ps', smoothscoreceil=3.4,
                    weight1='henikoff')

        self.assertIsInstance(aln0, egglib.Align)
        self.assertIsInstance(aln1, egglib.Align)
        self.assertIsInstance(aln2, egglib.Align)
        self.assertIsInstance(aln3, egglib.Align)
        self.assertIsInstance(aln4, egglib.Align)
        self.assertIsInstance(aln5, egglib.Align)

    def test_shortcut(self):
        egglib.wrappers.muscle(self.seq)
        egglib.wrappers.muscle3(self.seq)

    def test_muscle_E(self):
        cache = egglib.wrappers.paths['muscle']
        egglib.wrappers.paths['muscle']=None
        with self.assertRaises(RuntimeError):
            aln0=egglib.wrappers.muscle(self.seq, verbose=False)
        egglib.wrappers.paths['muscle'] = cache

        aln0=egglib.wrappers.muscle(self.seq, verbose=False)
        with self.assertRaises(TypeError):
            aln0=egglib.wrappers.muscle('dna', verbose=False)

        with self.assertRaises(TypeError):
            aln0=egglib.wrappers.muscle(self.seq, ref='error', verbose=False)

        with self.assertRaises(TypeError):
            egglib.wrappers.muscle(self.seq, ref=aln0, verbose=False)

        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(self.seq, verbose=False, error='error')

        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(self.seq, verbose=False, anchorspacing='error')

        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(self.seq, verbose=False, cluster1='error')

    def test_alphabet(self):
        dna_a = egglib.wrappers.muscle(self.seq ,verbose=False, seqtype='dna')
        prot_a = egglib.wrappers.muscle(self.prot ,verbose=False, seqtype='protein')
        with self.assertRaises(ValueError):
            codons_a = egglib.wrappers.muscle(self.codons ,verbose=False, seqtype='codon')
        with self.assertRaises(ValueError):
            codons_a = egglib.wrappers.muscle(self.codons ,verbose=False, seqtype='dna')
        egglib.wrappers.muscle(dna_a, ref=dna_a)
        egglib.wrappers.muscle(prot_a, ref=prot_a)
        egglib.wrappers.muscle(dna_a, ref=dna_a, seqtype='dna')
        egglib.wrappers.muscle(prot_a, ref=prot_a, seqtype='protein')

        with self.assertRaises(ValueError):
            prot_a = egglib.wrappers.muscle(self.prot ,verbose=False, seqtype='dna')
        with self.assertRaises(ValueError):
            dna_a = egglib.wrappers.muscle(self.seq ,verbose=False, seqtype='protein')

        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(dna_a, ref=prot_a)
        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(prot_a, ref=dna_a)

        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(dna_a, ref=prot_a, seqtype='dna')
        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(prot_a, ref=dna_a, seqtype='dna')

        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(prot_a, ref=dna_a, seqtype='protein')
        with self.assertRaises(ValueError):
            egglib.wrappers.muscle(dna_a, ref=prot_a, seqtype='protein')
